import datetime as dt
from unittest.mock import patch

import pook

from django.test import TestCase
from eveuniverse.tests.testdata.factories_2 import EveMoonFactory, EveTypeFactory

from structures.constants import EveCorporationId
from structures.models import OwnerCharacter, StarbaseDetail, Structure
from structures.tests.testdata.factories import (
    EveEntityCorporationFactory,
    OwnerFactory,
    PositionFactory,
    StarbaseFactory,
    StarbaseTypeFactory,
    UserMainDefaultOwnerFactory,
)

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
class TestUpdateStarbasesEsi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications

    @pook.on
    def test_can_sync_starbase_and_remove_old(self):
        # given
        fuel_1_type = EveTypeFactory()
        fuel_2_type = EveTypeFactory()
        moon = EveMoonFactory()
        starbase_type = StarbaseTypeFactory()
        starbase_id = 1300000000001
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": starbase_id,
                    "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=200,
            response_json=[
                {"item_id": starbase_id, "name": "Home Sweat Home"},
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        reinforced_until = dt.datetime(2020, 4, 5, 7, tzinfo=dt.timezone.utc)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=200,
            response_json=[
                {
                    "moon_id": moon.id,
                    "starbase_id": starbase_id,
                    "state": "online",
                    "system_id": moon.eve_planet.eve_solar_system.id,
                    "type_id": starbase_type.id,
                    "reinforced_until": reinforced_until.isoformat(),
                }
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases/{starbase_id}",
            reply=200,
            response_json={
                "allow_alliance_members": True,
                "allow_corporation_members": True,
                "anchor": "config_starbase_equipment_role",
                "attack_if_at_war": False,
                "attack_if_other_security_status_dropping": False,
                "fuel_bay_take": "config_starbase_equipment_role",
                "fuel_bay_view": "starbase_fuel_technician_role",
                "fuels": [
                    {
                        "quantity": 960,
                        "type_id": fuel_1_type.id,
                    },
                    {
                        "quantity": 11678,
                        "type_id": fuel_2_type.id,
                    },
                ],
                "offline": "config_starbase_equipment_role",
                "online": "config_starbase_equipment_role",
                "unanchor": "config_starbase_equipment_role",
                "use_alliance_standings": True,
            },
        )
        StarbaseFactory(owner=self.owner, id=1300000000099, name="delete-me")

        # when
        self.owner.update_structures_esi()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_structure_sync_fresh)

        # must contain all expected structures
        expected = {starbase_id}
        self.assertSetEqual(self.owner.structures.ids(), expected)

        # verify attributes for POS
        starbase = Structure.objects.get(id=starbase_id)
        self.assertEqual(starbase.name, "Home Sweat Home")
        self.assertEqual(
            starbase.eve_solar_system_id, moon.eve_planet.eve_solar_system.id
        )
        self.assertEqual(
            int(starbase.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(starbase.eve_type_id, starbase_type.id)
        self.assertEqual(starbase.state, Structure.State.POS_ONLINE)
        self.assertEqual(starbase.eve_moon_id, moon.id)
        self.assertEqual(starbase.state_timer_end, reinforced_until)
        self.assertEqual(starbase.position_x, 40.2)
        self.assertEqual(starbase.position_y, 27.3)
        self.assertEqual(starbase.position_z, -19.4)
        # verify details
        detail = starbase.starbase_detail
        self.assertTrue(detail.allow_alliance_members)
        self.assertTrue(detail.allow_corporation_members)
        self.assertEqual(
            detail.anchor_role, StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
        )
        self.assertFalse(detail.attack_if_at_war)
        self.assertFalse(detail.attack_if_other_security_status_dropping)
        self.assertEqual(
            detail.anchor_role, StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
        )
        self.assertEqual(
            detail.fuel_bay_take_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.fuel_bay_view_role,
            StarbaseDetail.Role.STARBASE_FUEL_TECHNICIAN_ROLE,
        )
        self.assertEqual(
            detail.offline_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.online_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.unanchor_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertTrue(detail.use_alliance_standings)
        # fuels
        self.assertEqual(detail.fuels.count(), 2)
        self.assertEqual(detail.fuels.get(eve_type_id=fuel_1_type.id).quantity, 960)
        self.assertEqual(detail.fuels.get(eve_type_id=fuel_2_type.id).quantity, 11678)

        # TODO: complete tests
        # structure = Structure.objects.get(id=1300000000002)
        # self.assertEqual(structure.name, "Bat cave")
        # self.assertEqual(structure.eve_solar_system_id, 30002537)
        # self.assertEqual(
        #     int(structure.owner.corporation.corporation_id), self.corporation_id
        # )
        # self.assertEqual(structure.eve_type_id, 20061)
        # self.assertEqual(structure.state, Structure.State.POS_OFFLINE)
        # self.assertEqual(structure.eve_moon_id, 40161466)
        # self.assertEqual(
        #     structure.unanchors_at, dt.datetime(2020, 5, 5, 7, 0, 0, tzinfo=dt.timezone.utc)
        # )
        # self.assertIsNone(structure.fuel_expires_at)
        # self.assertFalse(structure.generatednotification_set.exists())

        # structure = Structure.objects.get(id=1300000000003)
        # self.assertEqual(structure.name, "Panic Room")
        # self.assertEqual(structure.eve_solar_system_id, 30000474)
        # self.assertEqual(
        #     int(structure.owner.corporation.corporation_id), self.corporation_id
        # )
        # self.assertEqual(structure.eve_type_id, 20062)
        # self.assertEqual(structure.state, Structure.State.POS_REINFORCED)
        # self.assertEqual(structure.eve_moon_id, 40029527)
        # self.assertAlmostEqual(
        #     structure.fuel_expires_at,
        #     now() + dt.timedelta(seconds=360_000),
        #     delta=dt.timedelta(seconds=30),
        # )
        # self.assertEqual(
        #     structure.state_timer_end, dt.datetime(2020, 1, 2, 3, tzinfo=dt.timezone.utc)
        # )
        # self.assertTrue(structure.generatednotification_set.exists())

    @pook.on
    def test_can_update_starbase(self):
        # given
        starbase = StarbaseFactory(owner=self.owner)
        fuel_type = EveTypeFactory()
        reinforced_until = dt.datetime(2020, 4, 5, 7, tzinfo=dt.timezone.utc)

        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": starbase.id,
                    "position": PositionFactory(),
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=200,
            response_json=[
                {"item_id": starbase.id, "name": "Home Sweat Home"},
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=200,
            response_json=[
                {
                    "moon_id": starbase.eve_moon.id,
                    "starbase_id": starbase.id,
                    "state": "offline",
                    "system_id": starbase.eve_moon.eve_planet.eve_solar_system.id,
                    "type_id": starbase.eve_type.id,
                    "reinforced_until": reinforced_until.isoformat(),
                }
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases/{starbase.id}",
            reply=200,
            response_json={
                "allow_alliance_members": True,
                "allow_corporation_members": True,
                "anchor": "config_starbase_equipment_role",
                "attack_if_at_war": False,
                "attack_if_other_security_status_dropping": False,
                "fuel_bay_take": "config_starbase_equipment_role",
                "fuel_bay_view": "starbase_fuel_technician_role",
                "fuels": [
                    {
                        "quantity": 960,
                        "type_id": fuel_type.id,
                    },
                ],
                "offline": "config_starbase_equipment_role",
                "online": "config_starbase_equipment_role",
                "unanchor": "config_starbase_equipment_role",
                "use_alliance_standings": True,
            },
        )

        # when
        self.owner.update_structures_esi()

        # then
        self.assertEqual(self.owner.structures.count(), 1)
        starbase.refresh_from_db()
        # self.assertEqual(starbase.name, "Home Sweat Home")
        self.assertEqual(Structure.State(starbase.state), Structure.State.POS_OFFLINE)
        self.assertEqual(starbase.state_timer_end, reinforced_until)

    @patch(MODULE_PATH + ".STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".notify", spec=True)
    @pook.on
    def test_should_mark_error_when_character_not_director_while_updating_starbases(
        self, mock_notify
    ):
        # given
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=403,
            response_json={"error": "forbidden"},
        )

        # when
        self.owner.update_structures_esi()
        # then
        self.owner.refresh_from_db()
        self.assertFalse(self.owner.is_structure_sync_fresh)
        self.assertTrue(mock_notify)
        character: OwnerCharacter = self.owner.characters.first()
        self.assertEqual(character.error_count, 1)
        self.assertTrue(character.is_enabled)

    @patch(MODULE_PATH + ".STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".notify", spec=True)
    @pook.on
    def test_should_disable_character_when_not_director_while_updating_starbases(
        self, mock_notify
    ):
        # given
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=403,
            response_json={"error": "forbidden"},
        )

        character: OwnerCharacter = self.owner.characters.first()
        character.error_count = 3
        character.save()
        # when
        self.owner.update_structures_esi()
        # then
        self.owner.refresh_from_db()
        self.assertFalse(self.owner.is_structure_sync_fresh)
        self.assertTrue(mock_notify)
        character.refresh_from_db()
        self.assertFalse(character.is_enabled)

    @pook.on
    def test_should_reset_error_count_for_character_when_successful(self):
        # given
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=200,
            response_json=[],
        )

        character: OwnerCharacter = self.owner.characters.first()
        character.error_count = 3
        character.save()

        # when
        self.owner.update_structures_esi()
        # then

        character.refresh_from_db()
        self.assertTrue(character.is_enabled)
        self.assertEqual(character.error_count, 0)

    @pook.on
    def test_should_not_delete_existing_starbases_when_update_failed(self):
        # given
        starbase = StarbaseFactory(owner=self.owner)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=500,
            response_json={"error": "server error"},
        )

        # when
        self.owner.update_structures_esi()

        # then
        # self.assertFalse(self.owner.is_structure_sync_fresh)
        expected = {starbase.id}
        self.assertSetEqual(self.owner.structures.ids(), expected)

    @pook.on
    def test_should_not_break_when_starbase_names_not_found(self):
        # given
        moon = EveMoonFactory()
        starbase_type = StarbaseTypeFactory()
        starbase_id = 1300000000001
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": starbase_id,
                    "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=404,
            response_json={"error": "not found"},
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases",
            reply=200,
            response_json=[
                {
                    "moon_id": moon.id,
                    "starbase_id": starbase_id,
                    "state": "online",
                    "system_id": moon.eve_planet.eve_solar_system.id,
                    "type_id": starbase_type.id,
                    "reinforced_until": dt.datetime(
                        2020, 4, 5, 7, tzinfo=dt.timezone.utc
                    ).isoformat(),
                }
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/starbases/{starbase_id}",
            reply=200,
            response_json={
                "allow_alliance_members": True,
                "allow_corporation_members": True,
                "anchor": "config_starbase_equipment_role",
                "attack_if_at_war": False,
                "attack_if_other_security_status_dropping": False,
                "fuel_bay_take": "config_starbase_equipment_role",
                "fuel_bay_view": "starbase_fuel_technician_role",
                "fuels": [],
                "offline": "config_starbase_equipment_role",
                "online": "config_starbase_equipment_role",
                "unanchor": "config_starbase_equipment_role",
                "use_alliance_standings": True,
            },
        )

        # when
        self.owner.update_structures_esi()

        # then
        self.owner.refresh_from_db()
        expected = {starbase_id}
        self.assertSetEqual(self.owner.structures.ids(), expected)

    # Older below

    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    # @patch(MODULE_PATH + ".notify", spec=True)
    # def test_can_sync_all_structures_and_notify_user(self, mock_notify, mock_esi):
    #     # given
    #     mock_esi.client = self.esi_client_stub
    #     owner = OwnerFactory(user=self.user, structures_last_update_at=None)

    #     # when
    #     owner.update_structures_esi(user=self.user)

    #     # then
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)

    #     # must contain all expected structures
    #     expected = {
    #         1200000000003,
    #         1200000000004,
    #         1200000000005,
    #         1200000000006,
    #         1200000000099,
    #         1300000000001,
    #         1300000000002,
    #         1300000000003,
    #     }
    #     self.assertSetEqual(owner.structures.ids(), expected)

    #     # user report has been sent
    #     self.assertTrue(mock_notify.called)

    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    # def test_should_notify_admins_when_service_is_restored(
    #     self, mock_esi_client
    # ):
    #     # given
    #     mock_esi_client.side_effect = esi_mock_client
    #     owner = OwnerFactory(user=self.user, structures_last_update_at=None)
    #     owner.is_structure_sync_fresh = False
    #     owner.save()
    #     # when
    #     owner.update_structures_esi()
    #     # then
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)
    #     self.assertTrue(owner.is_structure_sync_fresh)
    #     self.assertTrue(owner.is_structure_sync_fresh)
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)
    #     self.assertTrue(owner.is_structure_sync_fresh)



"""ESI client provider for Structures."""

from pathlib import Path

from esi.openapi_clients import ESIClientProvider

from . import __version__

spec_file = Path(__file__).parent / "openapi_2025-12-16.json"
esi = ESIClientProvider(
    compatibility_date="2025-12-16",
    ua_appname="aa-structures",
    ua_version=__version__,
    operations=[
        "GetCharactersCharacterIdNotifications",
        "GetCorporationsCorporationIdAssets",
        "GetCorporationsCorporationIdCustomsOffices",
        "GetCorporationsCorporationIdStarbases",
        "GetCorporationsCorporationIdStarbasesStarbaseId",
        "GetCorporationsCorporationIdStructures",
        "GetSovereigntyMap",
        "GetUniverseStructuresStructureId",
        "GetUniverseStructuresStructureId",
        "PostCorporationsCorporationIdAssetsNames",
        "PostCorporationsCorporationIdAssetsLocations",
        "PostCorporationsCorporationIdAssetsNames",
    ],
    spec_file=spec_file,
)



import datetime as dt
import unicodedata

import pytz
from bs4 import BeautifulSoup
from markdown import markdown


def format_datetime_esi(my_dt: dt.datetime) -> str:
    """Convert datetime to ESI format, e.g `"2019-08-16T14:08:00Z"`."""
    utc_dt = my_dt.astimezone(pytz.utc)
    return utc_dt.strftime("%Y-%m-%dT%H:%M:%SZ")


def datetime_to_ldap(my_dt: dt.datetime) -> int:
    """Return a standard datetime as ldap datetime."""
    return (
        ((my_dt - dt.datetime(1970, 1, 1, tzinfo=pytz.utc)).total_seconds())
        + 11644473600
    ) * 10000000


def markdown_to_plain(text: str) -> str:
    """Convert text in markdown to plain text."""
    html = markdown(text)
    text = "".join(BeautifulSoup(html, features="html.parser").findAll(text=True))
    return unicodedata.normalize("NFKD", text)



from typing import NamedTuple, Optional

from eveuniverse.models import EveType

from app_utils.testing import NoSocketsTestCase

from structures.core import starbases
from structures.tests.testdata.factories import (
    CitadelTypeFactory,
    CustomsOfficeTypeFactory,
    StarbaseTypeFactory,
)


class TestStarbases(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.type_citadel = CitadelTypeFactory()
        cls.type_poco = CustomsOfficeTypeFactory()
        cls.type_starbase = StarbaseTypeFactory()

    def test_is_starbase(self):
        self.assertFalse(starbases.is_starbase(self.type_citadel))
        self.assertFalse(starbases.is_starbase(self.type_poco))
        self.assertTrue(starbases.is_starbase(self.type_starbase))

    def test_starbase_fuel_consumption_per_hour(self):
        class Case(NamedTuple):
            eve_type: EveType
            want: Optional[int]

        cases = [
            Case(
                StarbaseTypeFactory(id=16213, name="Caldari Control Tower"),
                40,
            ),
            Case(
                StarbaseTypeFactory(id=20061, name="Caldari Control Tower Medium"),
                20,
            ),
            Case(
                StarbaseTypeFactory(id=20062, name="Caldari Control Tower Small"),
                10,
            ),
            Case(CitadelTypeFactory(), None),
        ]

        for tc in cases:
            with self.subTest(name=tc.eve_type.name):
                self.assertEqual(starbases.fuel_per_hour(tc.eve_type), tc.want)

    def test_starbase_size(self):
        class Case(NamedTuple):
            eve_type: EveType
            want: Optional[starbases.StarbaseSize]

        cases = [
            Case(
                StarbaseTypeFactory(id=16213, name="Caldari Control Tower"),
                starbases.StarbaseSize.LARGE,
            ),
            Case(
                StarbaseTypeFactory(id=20061, name="Caldari Control Tower Medium"),
                starbases.StarbaseSize.MEDIUM,
            ),
            Case(
                StarbaseTypeFactory(id=20062, name="Caldari Control Tower Small"),
                starbases.StarbaseSize.SMALL,
            ),
            Case(CitadelTypeFactory(), None),
        ]

        for tc in cases:
            with self.subTest(name=tc.eve_type.name):
                self.assertEqual(starbases.starbase_size(tc.eve_type), tc.want)


class TestStarbasesFuelDuration(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def test_can_calculate_for_large_tower(self):
        # given
        large_tower_type = StarbaseTypeFactory(name="Caldari Control Tower")
        # when
        result = starbases.fuel_duration(
            starbase_type=large_tower_type, fuel_quantity=80
        )
        # then
        self.assertEqual(result, 7200)

    def test_can_calculate_for_large_tower_with_sov(self):
        # given
        large_tower_type = StarbaseTypeFactory(name="Caldari Control Tower")
        # when
        result = starbases.fuel_duration(
            starbase_type=large_tower_type, fuel_quantity=80, has_sov=True
        )
        # then
        self.assertEqual(result, 9600)

    def test_can_raise_error_when_not_starbase_type(self):
        # given
        astrahus_type = CitadelTypeFactory()
        # when
        with self.assertRaises(ValueError):
            starbases.fuel_duration(starbase_type=astrahus_type, fuel_quantity=80)
        # given
        astrahus_type = CitadelTypeFactory()
        # when
        with self.assertRaises(ValueError):
            starbases.fuel_duration(starbase_type=astrahus_type, fuel_quantity=80)
        # given
        astrahus_type = CitadelTypeFactory()
        # when
        with self.assertRaises(ValueError):
            starbases.fuel_duration(starbase_type=astrahus_type, fuel_quantity=80)
        with self.assertRaises(ValueError):
            starbases.fuel_duration(starbase_type=astrahus_type, fuel_quantity=80)
        with self.assertRaises(ValueError):
            starbases.fuel_duration(starbase_type=astrahus_type, fuel_quantity=80)



"""functions for loading test data and for building mocks"""

import datetime as dt
import json
import logging
from collections import namedtuple
from pathlib import Path
from random import randrange
from typing import Any, Dict, List

from django.forms.models import model_to_dict
from django.utils.timezone import now
from eveuniverse.tests.testdata.factories_2 import EveMoonFactory, EvePlanetFactory

from structures.core.notification_types import NotificationType
from structures.models import Notification, Owner
from structures.tests.testdata.factories import (
    CustomsOfficeFactory,
    EveEntityAllianceFactory,
    EveEntityCharacterFactory,
    EveEntityCorporationFactory,
    EveSolarSystemLowSecFactory,
    EveSolarSystemNullSecFactory,
    IHUBTypeFactory,
    MoonOreTypeFactory,
    RefineryFactory,
    SkyhookFactory,
    StarbaseFactory,
    StructureFactory,
    TCUTypeFactory,
)

from .factories import NotificationFactory

_current_folder = Path(__file__).parent
_FILENAME_EVEUNIVERSE_TESTDATA = "eveuniverse.json"

logger = logging.getLogger(__name__)


def test_data_filename():
    return f"{_current_folder}/{_FILENAME_EVEUNIVERSE_TESTDATA}"


##############################
# internal functions


def _load_notifications() -> List[Dict[str, Any]]:
    with (_current_folder / "notifications.json").open("r", encoding="utf-8") as f:
        _notifications = json.load(f)

    # update timestamp to current
    for notification in _notifications:
        notification["timestamp"] = now() - dt.timedelta(
            hours=randrange(3), minutes=randrange(60), seconds=randrange(60)
        )

    return _notifications


notifications = _load_notifications()


###################################
# helper functions
#


def load_notification_by_type(
    owner: Owner, notif_type: NotificationType
) -> Notification:
    for notification in notifications:
        if notification["type"] == notif_type.value:
            return NotificationFactory(
                owner=owner,
                notif_type=notif_type.value,
                text=notification.get("text", ""),
            )
    raise ValueError(f"Could not find notif for type: {notif_type}")


def load_notification_entities(owner: Owner):
    EveSolarSystemLowSecFactory(id=30002537, name="Amamake")
    EveSolarSystemNullSecFactory(id=30000474, name="1-PGSG")
    EvePlanetFactory(id=40161469, eve_type__id=2016, eve_solar_system__id=30000474)
    EveMoonFactory(id=40161465, eve_planet__id=40161469)
    EveMoonFactory(id=40161466, eve_planet__id=40161469)

    IHUBTypeFactory()
    MoonOreTypeFactory(id=46300)
    MoonOreTypeFactory(id=46301)
    MoonOreTypeFactory(id=46302)
    MoonOreTypeFactory(id=46303)
    TCUTypeFactory()

    EveEntityAllianceFactory(id=3001)
    EveEntityAllianceFactory(id=3002)
    EveEntityAllianceFactory(id=3011)
    EveEntityAllianceFactory(id=500012, name="Blood Raider Covenant")
    EveEntityCharacterFactory(id=1001)
    EveEntityCharacterFactory(id=1011)
    EveEntityCorporationFactory(id=1000134, name="Blood Raiders")
    EveEntityCorporationFactory(id=1000137, name="DED")
    EveEntityCorporationFactory(id=2001)
    EveEntityCorporationFactory(id=2002)
    EveEntityCorporationFactory(id=2011)
    EveEntityCorporationFactory(id=2021)
    EveEntityCorporationFactory(id=2901)
    EveEntityCorporationFactory(id=2902)

    CustomsOfficeFactory(
        eve_type__id=2233,
        eve_planet__id=40161469,
        owner=owner,
    )
    RefineryFactory(
        eve_moon__id=40161465,
        eve_type__id=35835,
        id=1000000000002,
        owner=owner,
    )
    StarbaseFactory(
        eve_moon__id=40161465,
        eve_type__id=16213,
        owner=owner,
    )
    StructureFactory(
        eve_solar_system__id=30002537,
        eve_type__id=35832,
        id=1000000000001,
        owner=owner,
    )
    SkyhookFactory(
        eve_planet__id=40161469,
        eve_type__id=81080,
        id=1000000010001,
        owner=owner,
    )


def load_notification_objects(owner: Owner, in_bulk=True):
    """Loads notification fixtures for an owner.

    Note that the notification require some entities to exit,
    which can be created with ``load_notification_entities()``.

    Args:
    - in_bulk: When disabled, will load notifications one by one (for debugging)
    """
    timestamp_start = now() - dt.timedelta(hours=2)
    objs = [
        _generate_notif_obj_for_owner(owner, timestamp_start, notification)
        for notification in notifications
    ]
    if in_bulk:
        Notification.objects.bulk_create(objs)
    else:
        for obj in objs:
            logger.info("Creating notif: %s", model_to_dict(obj))
            obj.save()


def _generate_notif_obj_for_owner(
    owner: Owner, timestamp_start: dt.datetime, notification: dict
):
    notification_id = notification["notification_id"]
    text = notification["text"] if "text" in notification else None
    is_read = notification["is_read"] if "is_read" in notification else None
    timestamp_start = timestamp_start + dt.timedelta(minutes=5)
    params = {
        "notification_id": notification_id,
        "owner": owner,
        "sender_id": notification["sender_id"],
        "timestamp": timestamp_start,
        "notif_type": notification["type"],
        "text": text,
        "is_read": is_read,
        "last_updated": now(),
        "is_sent": False,
    }
    return Notification(**params)


# def generate_eve_entities_from_auth_entities():
#     """Generate EveEntity objects from existing Auth EveOnline objects."""

#     def add_eve_entity(id, name, category):
#         if id not in existing_ids:
#             objs.append(EveEntity(id=id, category=category, name=name))
#             existing_ids.add(id)

#     existing_ids = set(EveEntity.objects.values_list("id", flat=True))
#     objs = []
#     for character in EveCharacter.objects.exclude(character_id__in=existing_ids):
#         add_eve_entity(
#             id=character.character_id,
#             name=character.character_name,
#             category=EveEntity.CATEGORY_CHARACTER,
#         )
#         add_eve_entity(
#             id=character.corporation_id,
#             name=character.corporation_name,
#             category=EveEntity.CATEGORY_CORPORATION,
#         )
#         if character.alliance_id:
#             add_eve_entity(
#                 id=character.alliance_id,
#                 name=character.alliance_name,
#                 category=EveEntity.CATEGORY_ALLIANCE,
#             )

#     for corporation in EveCorporationInfo.objects.exclude(
#         corporation_id__in=existing_ids
#     ):
#         add_eve_entity(
#             id=corporation.corporation_id,
#             name=corporation.corporation_name,
#             category=EveEntity.CATEGORY_CORPORATION,
#         )

#     for alliance in EveAllianceInfo.objects.exclude(alliance_id__in=existing_ids):
#         add_eve_entity(
#             id=alliance.alliance_id,
#             name=alliance.alliance_name,
#             category=EveEntity.CATEGORY_ALLIANCE,
#         )

#     EveEntity.objects.bulk_create(objs)


def clone_notification(obj: Notification) -> Notification:
    """Return clone of a Notification."""
    new_object = NotificationFactory(
        sender=obj.sender, notif_type=obj.notif_type, owner=obj.owner, text=obj.text
    )
    return new_object


NearestCelestial = namedtuple(
    "NearestCelestial", ["eve_type", "eve_object", "distance"]
)



from django.apps import AppConfig

from . import __version__


class StructuresConfig(AppConfig):
    name = "structures"
    label = "structures"
    verbose_name = f"Structures v{__version__}"

    def ready(self) -> None:
        from . import checks  # noqa: F401 pylint: disable=unused-import



from enum import IntEnum


class EveTypeId(IntEnum):
    """An EVE type ID."""

    ASTRAHUS = 35832
    ASTRAHUS_UPWELL_QUANTUM_CORE = 56201
    ATHANOR = 35835
    LIQUID_OZONE = 16273
    METENOX = 81826
    STANDUP_CLONING_CENTER_I = 35894
    STANDUP_METENOX_MOON_DRILL = 82941
    INTEGRATED_MOON_DRILL_ARMATURE = 81920


class EveMoonId(IntEnum):
    AMAMAKE_P2_M1 = 40161465


class EveSolarSystemId(IntEnum):
    AMAMAKE = 30002537



"""Creating or removing timers from notifications."""

# pylint: disable = ungrouped-imports

import datetime as dt

from django.utils.translation import gettext
from eveuniverse.models import EveType

from allianceauth.services.hooks import get_extension_logger
from app_utils.datetime import (
    DATETIME_FORMAT,
    ldap_time_2_datetime,
    ldap_timedelta_2_timedelta,
)
from app_utils.django import app_labels

from structures.app_settings import (
    STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED,
    STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
)
from structures.constants import EveTypeId
from structures.models import Notification, Structure

from . import sovereignty, starbases
from .notification_types import NotificationType

if "timerboard" in app_labels():
    from allianceauth.timerboard.models import Timer as AuthTimer
else:
    AuthTimer = None  # pylint: disable = invalid-name

if "structuretimers" in app_labels():
    from structuretimers.models import Timer
else:
    Timer = None  # pylint: disable = invalid-name

logger = get_extension_logger(__name__)


def add_or_remove_timer(notif: Notification) -> bool:
    """Add/remove a timer related to this notification for some types.

    Returns True when timers where added or removed, else False
    """
    match notif.notif_type:
        case (
            NotificationType.STRUCTURE_LOST_ARMOR
            | NotificationType.STRUCTURE_LOST_SHIELD
        ):
            timer_processed = _gen_timer_structure_reinforcement(notif)
        case NotificationType.SOV_STRUCTURE_REINFORCED:
            timer_processed = _gen_timer_sov_reinforcements(notif)
        case NotificationType.ORBITAL_REINFORCED:
            timer_processed = _gen_timer_customs_office_reinforcements(notif)
        case NotificationType.SKYHOOK_LOST_SHIELDS:
            timer_processed = _gen_timer_skyhook_reinforcements(notif)
        case (
            NotificationType.MOONMINING_EXTRACTION_STARTED
            | NotificationType.MOONMINING_EXTRACTION_CANCELLED
        ):
            if not STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED:
                timer_processed = None
            else:
                timer_processed = _gen_timer_moon_extraction(notif)
        case NotificationType.TOWER_REINFORCED_EXTRA:
            timer_processed = _gen_timer_tower_reinforcements(notif)
        case _:
            raise NotImplementedError(notif.notif_type)

    if timer_processed:
        logger.info("%s: Created timer for notification", notif.notification_id)
        notif.is_timer_added = True
        notif.save()

    return timer_processed


def _gen_timer_structure_reinforcement(notif: Notification) -> bool:
    """Generate timer for structure reinforcements"""
    token = notif.owner.fetch_token()
    parsed_text = notif.parsed_text()
    structure_obj, _ = Structure.objects.get_or_create_esi(
        id=parsed_text["structureID"], token=token
    )
    eve_time = notif.timestamp + ldap_timedelta_2_timedelta(parsed_text["timeLeft"])
    timer_processed = False
    if AuthTimer:
        details_map = {
            NotificationType.STRUCTURE_LOST_SHIELD: gettext("Armor timer"),
            NotificationType.STRUCTURE_LOST_ARMOR: gettext("Final timer"),
        }
        AuthTimer.objects.create(
            details=details_map.get(notif.notif_type, ""),
            system=structure_obj.eve_solar_system.name,
            planet_moon="",
            structure=structure_obj.eve_type.name,
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if Timer:
        timer_map = {
            NotificationType.STRUCTURE_LOST_SHIELD: Timer.Type.ARMOR,
            NotificationType.STRUCTURE_LOST_ARMOR: Timer.Type.HULL,
        }
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=structure_obj.eve_solar_system,
            structure_type=structure_obj.eve_type,
            timer_type=timer_map.get(notif.notif_type),
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name=structure_obj.name,
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_sov_reinforcements(notif: Notification) -> bool:
    """Generate timer for sov reinforcements."""
    if not notif.owner.is_alliance_main:
        return False

    parsed_text = notif.parsed_text()
    solar_system = notif.eve_solar_system()
    structure_type_name = sovereignty.event_type_to_structure_type_name(
        parsed_text["campaignEventType"]
    )
    eve_time = ldap_time_2_datetime(parsed_text["decloakTime"])
    timer_processed = False

    if AuthTimer:
        AuthTimer.objects.create(
            details=gettext("Sov timer"),
            system=solar_system.name,
            planet_moon="",
            structure=structure_type_name,
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if Timer:
        structure_type, _ = EveType.objects.get_or_create_esi(
            id=sovereignty.event_type_to_type_id(parsed_text["campaignEventType"])
        )
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=solar_system,
            structure_type=structure_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            owner_name=notif.sender.name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_customs_office_reinforcements(notif: Notification) -> bool:
    """Generate timer for customs office reinforcements."""

    solar_system = notif.eve_solar_system()
    planet = notif.eve_planet()
    parsed_text = notif.parsed_text()
    eve_time = ldap_time_2_datetime(parsed_text["reinforceExitTime"])
    timer_processed = False

    if AuthTimer:
        AuthTimer.objects.create(
            details=gettext("Final timer"),
            system=solar_system.name,
            planet_moon=planet.name,
            structure="POCO",
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if Timer:
        structure_type, _ = EveType.objects.get_or_create_esi(
            id=EveTypeId.CUSTOMS_OFFICE
        )
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=solar_system,
            structure_type=structure_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            location_details=planet.name,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name="Customs Office",
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_skyhook_reinforcements(notif: Notification) -> bool:
    """Generate timer for skyhook reinforcements."""

    solar_system = notif.eve_solar_system("solarsystemID")
    structure_type = notif.eve_structure_type("typeID")
    planet = notif.eve_planet()
    eve_time = ldap_time_2_datetime(notif.parsed_text()["timestamp"])
    timer_processed = False

    if AuthTimer:
        AuthTimer.objects.create(
            details=gettext("Final timer"),
            system=solar_system.name,
            planet_moon=planet.name,
            structure="POCO",  # Auth timer does not support the Skyhook type yet
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if Timer:
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=solar_system,
            structure_type=structure_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            location_details=planet.name,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name=structure_type.name,
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _gen_timer_moon_extraction(notif: Notification) -> bool:
    """Generate timer for moon mining extractions."""
    solar_system = notif.eve_solar_system()
    moon = notif.eve_moon()
    parsed_text = notif.parsed_text()
    eve_time = _extract_eve_time(parsed_text)

    structure_type_name = "Moon Mining Cycle"
    objective = "Friendly"
    timer_processed = False
    structure_type = notif.eve_structure_type()

    match notif.notif_type:
        case NotificationType.MOONMINING_EXTRACTION_STARTED:
            if AuthTimer:
                AuthTimer.objects.create(
                    details=gettext("Extraction ready"),
                    system=solar_system.name,
                    planet_moon=moon.name,
                    structure=structure_type_name,
                    objective=objective,
                    eve_time=eve_time,
                    eve_corp=notif.owner.corporation,
                    corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
                )
                timer_processed = True

            if Timer:
                Timer.objects.create(
                    eve_solar_system=solar_system,
                    structure_type=structure_type,
                    timer_type=Timer.Type.MOONMINING,
                    objective=Timer.Objective.FRIENDLY,
                    date=eve_time,
                    location_details=moon.name,
                    eve_corporation=notif.owner.corporation,
                    eve_alliance=notif.owner.corporation.alliance,
                    visibility=_calc_visibility(),
                    structure_name=parsed_text["structureName"],
                    owner_name=notif.owner.corporation.corporation_name,
                    details_notes=_timer_details_notes(notif),
                )
                timer_processed = True

        case NotificationType.MOONMINING_EXTRACTION_CANCELLED:
            timer_processed = True

            for notification in Notification.objects.filter(
                notif_type=NotificationType.MOONMINING_EXTRACTION_STARTED,
                owner=notif.owner,
                is_timer_added=True,
                timestamp__lte=notif.timestamp,
            ).order_by("-timestamp"):
                notification: Notification
                parsed_text_2 = notification.parsed_text()
                my_structure_type_id = parsed_text_2["structureTypeID"]
                if my_structure_type_id == parsed_text["structureTypeID"]:
                    eve_time_2 = _extract_eve_time(parsed_text_2)

                    if AuthTimer:
                        timer_query = AuthTimer.objects.filter(
                            system=solar_system.name,
                            planet_moon=moon.name,
                            structure=structure_type_name,
                            objective=objective,
                            eve_time=eve_time_2,
                        )
                        deleted_count, _ = timer_query.delete()
                        logger.info(
                            "%s: removed %d obsolete Auth timers related to notification",
                            notif.notification_id,
                            deleted_count,
                        )

                    if Timer:
                        timer_query = Timer.objects.filter(
                            eve_solar_system=solar_system,
                            structure_type=structure_type,
                            timer_type=Timer.Type.MOONMINING,
                            location_details=moon.name,
                            date=eve_time_2,
                            objective=Timer.Objective.FRIENDLY,
                            eve_corporation=notif.owner.corporation,
                            eve_alliance=notif.owner.corporation.alliance,
                            visibility=_calc_visibility(),
                            structure_name=parsed_text["structureName"],
                            owner_name=notif.owner.corporation.corporation_name,
                        )
                        deleted_count, _ = timer_query.delete()
                        logger.info(
                            "%s: removed %d obsolete structure timers "
                            "related to notification",
                            notif.notification_id,
                            deleted_count,
                        )

    return timer_processed


def _calc_visibility():
    if Timer:
        return (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
    return None


def _extract_eve_time(parsed_text):
    if "readyTime" in parsed_text:
        return ldap_time_2_datetime(parsed_text["readyTime"])
    return None


def _gen_timer_tower_reinforcements(notif: Notification) -> bool:
    """Generate timer for tower reinforcements."""
    structure = notif.structures.select_related(
        "eve_solar_system", "eve_type", "eve_moon"
    ).first()
    eve_time = dt.datetime.fromisoformat(notif.details["reinforced_until"])
    timer_processed = False

    if AuthTimer:
        structure_type_map = {
            starbases.StarbaseSize.SMALL: "POS[S]",
            starbases.StarbaseSize.MEDIUM: "POS[M]",
            starbases.StarbaseSize.LARGE: "POS[L]",
        }
        structure_str = structure_type_map.get(
            starbases.starbase_size(structure.eve_type), "POS[M]"
        )
        AuthTimer.objects.create(
            details=gettext("Final timer"),
            system=structure.eve_solar_system.name,
            planet_moon=structure.eve_moon.name,
            structure=structure_str,
            objective="Friendly",
            eve_time=eve_time,
            eve_corp=notif.owner.corporation,
            corp_timer=STRUCTURES_TIMERS_ARE_CORP_RESTRICTED,
        )
        timer_processed = True

    if Timer:
        visibility = (
            Timer.Visibility.CORPORATION
            if STRUCTURES_TIMERS_ARE_CORP_RESTRICTED
            else Timer.Visibility.UNRESTRICTED
        )
        Timer.objects.create(
            eve_solar_system=structure.eve_solar_system,
            structure_type=structure.eve_type,
            timer_type=Timer.Type.FINAL,
            objective=Timer.Objective.FRIENDLY,
            date=eve_time,
            location_details=structure.eve_moon.name,
            eve_corporation=notif.owner.corporation,
            eve_alliance=notif.owner.corporation.alliance,
            visibility=visibility,
            structure_name=structure.name,
            owner_name=notif.owner.corporation.corporation_name,
            details_notes=_timer_details_notes(notif),
        )
        timer_processed = True

    return timer_processed


def _timer_details_notes(notif: Notification) -> str:
    """Return generated details notes string for Timers."""
    return (
        "Automatically created from structure notification for "
        f"{notif.owner.corporation} at {notif.timestamp.strftime(DATETIME_FORMAT)}"
    )



import datetime as dt
from unittest.mock import patch

import pook

from django.test import TestCase
from django.utils.timezone import now
from eveuniverse.tests.testdata.factories_2 import EveMoonFactory, EveSolarSystemFactory

from structures.constants import EveCorporationId
from structures.core.notification_types import NotificationType
from structures.models import Structure, StructureService
from structures.tests.testdata.factories import (
    CitadelTypeFactory,
    EveEntityCorporationFactory,
    FuelAlertConfigFactory,
    MoonDrillTypeFactory,
    OwnerFactory,
    StructureFactory,
    StructureTagFactory,
    UserMainDefaultOwnerFactory,
    WebhookFactory,
)
from structures.tests.testdata.helpers import NearestCelestial

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
class TestUpdateStructuresEsi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id

    @pook.on
    def test_can_create_new_structures_and_delete_old(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        StructureFactory(owner=owner, id=1000000000004, name="delete-me")

        structure_id = 1000000000001
        reinforce_hour = 18
        next_reinforce_hour = 12
        state_timer_start = now() + dt.timedelta(hours=3)
        state_timer_end = state_timer_start + dt.timedelta(days=2, hours=1)
        unanchors_at = now() + dt.timedelta(hours=12)
        fuel_expires_at = now() + dt.timedelta(days=3)
        solar_system = EveSolarSystemFactory(name="Amamake")
        structure_type = CitadelTypeFactory()
        next_reinforce_apply = now() + dt.timedelta(days=5)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "fuel_expires": fuel_expires_at.isoformat(),
                    "next_reinforce_apply": next_reinforce_apply.isoformat(),
                    "next_reinforce_hour": next_reinforce_hour,
                    "profile_id": 101853,
                    "reinforce_hour": reinforce_hour,
                    "services": [
                        {"name": "Clone Bay", "state": "online"},
                        {"name": "Market Hub", "state": "offline"},
                    ],
                    "state": "shield_vulnerable",
                    "state_timer_end": state_timer_end.isoformat(),
                    "state_timer_start": state_timer_start.isoformat(),
                    "structure_id": structure_id,
                    "system_id": solar_system.id,
                    "type_id": structure_type.id,
                    "unanchors_at": unanchors_at.isoformat(),
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure_id}",
            reply=200,
            response_json={
                "owner_id": self.corporation_id,
                "name": f"{solar_system.name} - Test Structure Alpha",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": solar_system.id,
                "type_id": structure_type.id,
            },
        )

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertAlmostEqual(
            owner.structures_last_update_at, now(), delta=dt.timedelta(seconds=30)
        )
        self.assertSetEqual(owner.structures.ids(), {1000000000001})

        # verify attributes for structure
        structure = Structure.objects.get(id=1000000000001)
        self.assertEqual(structure.name, "Test Structure Alpha")
        self.assertEqual(structure.position_x, 55028384780.0)
        self.assertEqual(structure.position_y, 7310316270.0)
        self.assertEqual(structure.position_z, -163686684205.0)
        self.assertEqual(structure.eve_solar_system_id, solar_system.id)
        self.assertEqual(structure.eve_type_id, structure_type.id)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.state, Structure.State.SHIELD_VULNERABLE)
        self.assertEqual(structure.reinforce_hour, reinforce_hour)
        self.assertEqual(structure.fuel_expires_at, fuel_expires_at)
        self.assertEqual(structure.state_timer_start, state_timer_start)
        self.assertEqual(structure.state_timer_end, state_timer_end)
        self.assertEqual(structure.unanchors_at, unanchors_at)
        self.assertEqual(structure.next_reinforce_apply, next_reinforce_apply)
        self.assertEqual(structure.next_reinforce_hour, next_reinforce_hour)
        want = {
            "Clone Bay": StructureService.State.ONLINE,
            "Market Hub": StructureService.State.OFFLINE,
        }
        got = {x.name: x.state for x in structure.services.all()}
        self.assertDictEqual(got, want)

    @pook.on
    def test_can_update_existing_structure(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure = StructureFactory(
            owner=owner,
            state=Structure.State.ANCHOR_VULNERABLE,
        )
        reinforce_hour = 18
        next_reinforce_hour = 12
        state_timer_start = now() + dt.timedelta(hours=3)
        state_timer_end = state_timer_start + dt.timedelta(days=2, hours=1)
        unanchors_at = now() + dt.timedelta(hours=12)
        fuel_expires_at = now() + dt.timedelta(days=3)
        next_reinforce_apply = now() + dt.timedelta(days=5)

        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "fuel_expires": fuel_expires_at.isoformat(),
                    "next_reinforce_apply": next_reinforce_apply.isoformat(),
                    "next_reinforce_hour": next_reinforce_hour,
                    "profile_id": 101853,
                    "reinforce_hour": reinforce_hour,
                    "services": [
                        {"name": "Clone Bay", "state": "online"},
                        {"name": "Market Hub", "state": "offline"},
                    ],
                    "state": "shield_vulnerable",
                    "state_timer_end": state_timer_end.isoformat(),
                    "state_timer_start": state_timer_start.isoformat(),
                    "structure_id": structure.id,
                    "system_id": structure.eve_solar_system.id,
                    "type_id": structure.eve_type.id,
                    "unanchors_at": unanchors_at.isoformat(),
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure.id}",
            reply=200,
            response_json={
                "owner_id": self.corporation_id,
                "name": f"{structure.eve_solar_system.name} - {structure.name}",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": structure.eve_solar_system.id,
                "type_id": structure.eve_type.id,
            },
        )

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertAlmostEqual(
            owner.structures_last_update_at, now(), delta=dt.timedelta(seconds=30)
        )
        self.assertSetEqual(owner.structures.ids(), {structure.id})

        # verify attributes for structure
        structure.refresh_from_db()
        self.assertEqual(structure.state, Structure.State.SHIELD_VULNERABLE)
        self.assertEqual(structure.reinforce_hour, reinforce_hour)
        self.assertEqual(structure.fuel_expires_at, fuel_expires_at)
        self.assertEqual(structure.state_timer_start, state_timer_start)
        self.assertEqual(structure.state_timer_end, state_timer_end)
        self.assertEqual(structure.unanchors_at, unanchors_at)
        self.assertEqual(structure.next_reinforce_apply, next_reinforce_apply)
        self.assertEqual(structure.next_reinforce_hour, next_reinforce_hour)
        want = {
            "Clone Bay": StructureService.State.ONLINE,
            "Market Hub": StructureService.State.OFFLINE,
        }
        got = {x.name: x.state for x in structure.services.all()}
        self.assertDictEqual(got, want)

    @pook.on
    def test_can_handle_owner_without_structures(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertSetEqual(owner.structures.ids(), set())

    @pook.on
    def test_should_not_break_when_structures_endpoint_returns_error(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=500,
            response_json={"error": "internal error"},
        )

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        self.assertSetEqual(owner.structures.ids(), set())

    @pook.on
    def test_should_not_delete_existing_upwell_structures_when_update_failed(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure = StructureFactory(owner=owner)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=500,
            response_json={"error": "internal error"},
        )
        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        self.assertSetEqual(owner.structures.ids(), {structure.id})

    @pook.on
    def test_should_not_break_when_universe_endpoint_returns_error(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure_id = 1000000000001
        solar_system = EveSolarSystemFactory()
        structure_type = CitadelTypeFactory()
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure_id,
                    "system_id": solar_system.id,
                    "type_id": structure_type.id,
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure_id}",
            reply=500,
            response_json={"error": "internal_error"},
        )

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=structure_id)
        self.assertEqual(structure.name, "(no data)")

    @pook.on
    def test_update_will_not_break_on_403_error_from_structure_info(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure_id = 1000000000001
        solar_system = EveSolarSystemFactory()
        structure_type = CitadelTypeFactory()
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure_id,
                    "system_id": solar_system.id,
                    "type_id": structure_type.id,
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure_id}",
            reply=403,
            response_json={"error": "forbidden"},
        )

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=structure_id)
        self.assertEqual(structure.name, "(no data)")

    @pook.on
    def test_tags_are_not_modified_by_update(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure = StructureFactory(owner=owner)

        tag_a = StructureTagFactory(name="tag_a")
        structure.tags.add(tag_a)
        structure.save()

        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure.id,
                    "system_id": structure.eve_solar_system.id,
                    "type_id": structure.eve_type.id,
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure.id}",
            reply=200,
            response_json={
                "owner_id": self.corporation_id,
                "name": f"{structure.eve_solar_system.name} - Test Structure Alpha",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": structure.eve_solar_system.id,
                "type_id": structure.eve_type.id,
            },
        )

        # when
        owner.update_structures_esi()

        # then
        self.assertSetEqual(owner.structures.ids(), {structure.id})

        # should still contain the tag
        self.assertTrue(structure.tags.filter(name="tag_a").exists())


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
class TestUpdateStructuresEsi_FuelAlerts(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications

    @patch(
        "structures.models.structures_1.STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
    )
    @patch("structures.models.notifications.Webhook.send_message")
    @pook.on
    def test_should_send_refueled_notification_when_fuel_level_increased(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        owner.webhooks.add(webhook)

        fuel_expires_at_1 = now() + dt.timedelta(hours=3)
        fuel_expires_at_2 = fuel_expires_at_1 + dt.timedelta(days=3)
        structure = StructureFactory(owner=owner, fuel_expires_at=fuel_expires_at_1)
        config = FuelAlertConfigFactory(start=48, end=0, repeat=12)
        structure.structure_fuel_alerts.create(config=config, hours=12)
        structure.save()

        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "fuel_expires": fuel_expires_at_2.isoformat(),
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure.id,
                    "system_id": structure.eve_solar_system.id,
                    "type_id": structure.eve_type.id,
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure.id}",
            reply=200,
            response_json={
                "owner_id": self.corporation_id,
                "name": f"{structure.eve_solar_system.name} - Test Structure Alpha",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": structure.eve_solar_system.id,
                "type_id": structure.eve_type.id,
            },
        )

        # when
        owner.update_structures_esi()

        # then
        self.assertTrue(mock_send_message.called)
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    @patch(
        "structures.models.structures_1.STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
    )
    @patch("structures.models.notifications.Webhook.send_message")
    @pook.on
    def test_should_not_send_refueled_notification_when_fuel_level_unchanged(
        self, mock_send_message
    ):
        # given
        mock_send_message.return_value = 1
        structure_id = 1000000000001
        fuel_expires_at_1 = now() + dt.timedelta(hours=3)
        fuel_expires_at_2 = fuel_expires_at_1
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        owner.webhooks.add(webhook)
        structure = StructureFactory(
            owner=owner, id=structure_id, fuel_expires_at=fuel_expires_at_1
        )

        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "fuel_expires": fuel_expires_at_2.isoformat(),
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure_id,
                    "system_id": structure.eve_solar_system.id,
                    "type_id": structure.eve_type.id,
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure_id}",
            reply=200,
            response_json={
                "owner_id": self.corporation_id,
                "name": f"{structure.eve_solar_system.name} - Test Structure Alpha",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": structure.eve_solar_system.id,
                "type_id": structure.eve_type.id,
            },
        )

        # when
        owner.update_structures_esi()

        # then
        self.assertFalse(mock_send_message.called)

    @pook.on
    def test_can_update_metenox(self):
        # given
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        structure_id = 1000000000111
        structure_name = "Test Moon Drill"
        solar_system = EveSolarSystemFactory(name="Amamake")
        structure_type = MoonDrillTypeFactory(id=81826, name="Metenox Moon Drill")

        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": self.corporation_id,
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure_id,
                    "system_id": solar_system.id,
                    "type_id": structure_type.id,
                    "services": [
                        {"name": "Moon Drill", "state": "online"},
                    ],
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure_id}",
            reply=200,
            response_json={
                "owner_id": self.corporation_id,
                "name": f"{solar_system.name} - {structure_name}",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": solar_system.id,
                "type_id": structure_type.id,
            },
        )

        moon = EveMoonFactory(eve_planet__eve_solar_system=solar_system)

        # when
        with patch(MODULE_PATH + ".EveSolarSystem.nearest_celestial") as m:
            m.return_value = NearestCelestial(None, moon, 100)
            owner.update_structures_esi()

        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        structure = Structure.objects.get(id=structure_id)
        self.assertEqual(structure.name, structure_name)
        self.assertEqual(structure.eve_solar_system.id, solar_system.id)
        services = set(structure.services.values_list("name", flat=True))
        self.assertSetEqual({"Moon Drill"}, services)
        self.assertEqual(structure.eve_moon, moon)



"""Constants for Structures."""

from enum import IntEnum


# Eve enums
class EveAttributeId(IntEnum):
    """An EVE attribute ID."""

    SQUADRON_SIZE = 2215


class EveCategoryId(IntEnum):
    """An EVE category ID."""

    ORBITAL = 46
    STARBASE = 23
    STRUCTURE = 65
    STRUCTURE_MODULE = 66


class EveGroupId(IntEnum):
    """An EVE group ID."""

    CITADEL = 1657
    CONTROL_TOWER = 365
    ENGINEERING_COMPLEX = 1404
    FUEL_BLOCK = 1136
    ICE_PRODUCT = 423
    MOON = 8
    PLANET = 7
    QUANTUM_CORES = 4086
    REFINERY = 1406
    STRUCTURE_CITADEL_SERVICE_MODULE = 1321
    UNCOMMON_MOON_ASTEROIDS = 1921
    UPWELL_JUMP_BRIDGE = 1408
    UPWELL_MOON_DRILL = 4744


class EveTypeId(IntEnum):
    """An EVE type ID."""

    ANSIBLE_JUMP_BRIDGE = 35841
    ASTRAHUS_UPWELL_QUANTUM_CORE = 56201
    CALDARI_CONTROL_TOWER = 16213
    CUSTOMS_OFFICE = 2233
    IHUB = 32458
    LIQUID_OZONE = 16273
    NITROGEN_FUEL_BLOCK = 4051
    ORBITAL_SKYHOOK = 81080
    STRONTIUM = 16275
    TCU = 32226


class EveCorporationId(IntEnum):
    """An EVE corporation ID."""

    DED = 1000137



import datetime as dt
from unittest.mock import patch

import pook
import yaml

from django.test import TestCase, override_settings
from django.utils.timezone import now
from esi.exceptions import HTTPServerError
from esi.models import Token
from eveuniverse.tests.testdata.factories_2 import EveMoonFactory, EvePlanetFactory

from app_utils.testing import NoSocketsTestCase, queryset_pks

from structures.core.notification_types import NotificationType
from structures.models import Notification, Structure, StructureItem
from structures.tests.helpers import datetime_to_ldap
from structures.tests.testdata.factories import (
    CitadelServiceModuleTypeFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
    EveEntityCorporationFactory,
    JumpFuelAlertConfigFactory,
    LiquidOzoneTypeFactory,
    MoonOreTypeFactory,
    OwnerFactory,
    PositionFactory,
    QuantumCoreTypeFactory,
    RefineryFactory,
    SkyhookFactory,
    SkyhookTypeFactory,
    StarbaseFactory,
    StructureFactory,
    StructureItemFactory,
    UserMainDefaultOwnerFactory,
    WebhookFactory,
)
from structures.tests.testdata.helpers import (
    NearestCelestial,
    load_notification_entities,
    load_notification_objects,
)

OWNERS_PATH = "structures.models.owners"
NOTIFICATIONS_PATH = "structures.models.notifications"


class TestFetchNotificationsEsi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.sender_corporation = EveEntityCorporationFactory()
        cls.character = EveCharacterFactory()
        cls.user = UserMainDefaultOwnerFactory(main_character__character=cls.character)

    @pook.on
    def test_should_create_notifications_from_scratch(self):
        # given
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        notification_id = 1000000916
        recruit = EveCharacterFactory()
        timestamp = now()
        data = {
            "charID": recruit.character_id,
            "corpID": recruit.corporation_id,
        }
        text = yaml.dump(data)
        pook.get(
            f"https://esi.evetech.net/characters/{self.character.character_id}/notifications",
            reply=200,
            response_json=[
                {
                    "notification_id": notification_id,
                    "type": "CharLeftCorpMsg",
                    "sender_id": self.sender_corporation.id,
                    "sender_type": "corporation",
                    "timestamp": timestamp.isoformat(),
                    "text": text,
                    "is_read": True,
                },
            ],
        )

        # when
        owner.fetch_notifications_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_notification_sync_fresh)
        self.assertEqual(owner.notification_set.count(), 1)
        notif: Notification = owner.notification_set.get(
            notification_id=notification_id
        )
        self.assertEqual(notif.notification_id, notification_id)
        self.assertEqual(notif.notif_type, NotificationType.CHAR_LEFT_CORP_MSG)
        self.assertEqual(notif.sender, self.sender_corporation)
        self.assertEqual(notif.timestamp, timestamp)
        self.assertEqual(notif.text, text)
        self.assertTrue(notif.is_read)
        self.assertFalse(notif.is_sent)
        self.assertEqual(notif.parsed_text(), data)

    @pook.on
    def test_should_handle_other_sender_correctly(self):
        # given
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        notification_id = 1000000916
        pook.get(
            f"https://esi.evetech.net/characters/{self.character.character_id}/notifications",
            reply=200,
            response_json=[
                {
                    "notification_id": notification_id,
                    "type": "CorpBecameWarEligible",
                    "sender_id": 42,
                    "sender_type": "other",
                    "timestamp": now().isoformat(),
                    "text": "",
                },
            ],
        )

        # when
        owner.fetch_notifications_esi()

        # then
        obj = owner.notification_set.get(notification_id=notification_id)
        self.assertIsNone(obj.sender)

    @patch(OWNERS_PATH + ".notify", spec=True)
    @pook.on
    def test_should_inform_user_about_successful_update(self, mock_notify):
        # given
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        notification_id = 1000000916
        pook.get(
            f"https://esi.evetech.net/characters/{self.character.character_id}/notifications",
            reply=200,
            response_json=[
                {
                    "notification_id": notification_id,
                    "type": "CorpBecameWarEligible",
                    "sender_id": self.sender_corporation.id,
                    "sender_type": "corporation",
                    "timestamp": now().isoformat(),
                    "text": "",
                },
            ],
        )

        # when
        owner.fetch_notifications_esi(self.user)

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_notification_sync_fresh)
        self.assertTrue(mock_notify.called)

    @pook.on
    def test_should_set_moon_for_structure_when_missing(self):
        # given
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        notification_id = 1000000404
        ore_1 = MoonOreTypeFactory()
        ore_2 = MoonOreTypeFactory()
        ore_3 = MoonOreTypeFactory()
        refinery = RefineryFactory(owner=owner, eve_moon=None)
        moon = EveMoonFactory(eve_planet__eve_solar_system=refinery.eve_solar_system)
        timestamp = now()
        ready_time = timestamp + dt.timedelta(days=3)
        data = {
            "autoTime": datetime_to_ldap(ready_time + dt.timedelta(hours=2)),
            "moonID": moon.id,
            "oreVolumeByType": {
                ore_1.id: 1288475.124715103,
                ore_2.id: 544691.7637724016,
                ore_3.id: 526825.4047522942,
            },
            "readyTime": datetime_to_ldap(ready_time),
            "solarSystemID": refinery.eve_solar_system.id,
            "startedBy": self.character.character_id,
            "startedByLink": (
                f'<a href="showinfo:1383//{self.character.character_id}">'
                f"{self.character.character_name}</a>"
            ),
            "structureID": refinery.id,
            "structureLink": (
                f'<a href="showinfo:{refinery.eve_type.id}//{refinery.id}">'
                f"{refinery.name}</a>"
            ),
            "structureName": "Dummy",
            "structureTypeID": refinery.eve_type.id,
        }
        pook.get(
            f"https://esi.evetech.net/characters/{self.character.character_id}/notifications",
            reply=200,
            response_json=[
                {
                    "notification_id": notification_id,
                    "type": "MoonminingExtractionStarted",
                    "sender_id": self.sender_corporation.id,
                    "sender_type": "corporation",
                    "timestamp": timestamp.isoformat(),
                    "text": yaml.dump(data),
                },
            ],
        )

        # when
        owner.fetch_notifications_esi()

        # then
        refinery.refresh_from_db()
        self.assertEqual(refinery.eve_moon_id, moon.id)

    @pook.on
    def test_should_bubble_up_http_error_during_sync(self):
        owner = OwnerFactory(
            user=self.user,
            notifications_last_update_at=None,
            characters=[self.character],
        )
        pook.get(
            f"https://esi.evetech.net/characters/{self.character.character_id}/notifications",
            reply=502,
            response_json={"error": "some error"},
        )
        # when
        with self.assertRaises(HTTPServerError):
            owner.fetch_notifications_esi()

        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_notification_sync_fresh)


@override_settings(DEBUG=True)
@patch(NOTIFICATIONS_PATH + ".STRUCTURES_REPORT_NPC_ATTACKS", True)
@patch(NOTIFICATIONS_PATH + ".Webhook.send_message", spec=True)
class TestSendNewNotifications(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.owner = OwnerFactory(
            is_alliance_main=True, webhooks=False, forwarding_last_update_at=None
        )
        load_notification_entities(cls.owner)
        load_notification_objects(cls.owner)

    def setUp(self) -> None:
        self.owner.webhooks.clear()

    # TODO: Temporarily disabled
    # @patch(
    #     NOTIFICATIONS_PATH + ".STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS", False
    # )
    def test_should_send_all_notifications(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        webhook = WebhookFactory(notification_types=NotificationType.values)
        self.owner.webhooks.add(webhook)

        # when
        self.owner.send_new_notifications()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_forwarding_sync_fresh)
        notifications_processed = {
            int(args[1]["embeds"][0].footer.text[-10:])
            for args in mock_send_message.call_args_list
        }
        notifications_expected = set(
            self.owner.notification_set.filter(
                notif_type__in=NotificationType.values
            ).values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notifications_processed, notifications_expected)

    # TODO: temporary disabled
    # @patch(
    #     NOTIFICATIONS_PATH + ".STRUCTURES_NOTIFICATION_DISABLE_ESI_FUEL_ALERTS", True
    # )
    # def test_should_send_all_notifications_except_fuel_alerts(self, mock_send_message):
    #     # given
    #     mock_send_message.return_value = True
    #     self.user = AuthUtils.add_permission_to_user_by_name(
    #         "structures.add_structure_owner", self.user
    #     )
    #     # when
    #     self.self.owner.send_new_notifications()
    #     # then
    #     self.self.owner.refresh_from_db()
    #     self.assertTrue(self.self.owner.is_forwarding_sync_fresh)
    #     notifications_processed = {
    #         int(args[1]["embeds"][0].footer.text[-10:])
    #         for args in mock_send_message.call_args_list
    #     }
    #     notif_types = set(NotificationType.values)
    #     notif_types.discard(NotificationType.STRUCTURE_FUEL_ALERT)
    #     notif_types.discard(NotificationType.TOWER_RESOURCE_ALERT_MSG)
    #     notifications_expected = set(
    #         self.self.owner.notifications.filter(notif_type__in=notif_types).values_list(
    #             "notification_id", flat=True
    #         )
    #     )
    #     self.assertSetEqual(notifications_processed, notifications_expected)

    def test_should_send_all_notifications_corp(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        corporation = EveCorporationInfoFactory(corporation_id=2011)
        character = EveCharacterFactory(corporation=corporation)
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(
            user=user,
            is_alliance_main=True,
            webhooks__notification_types=NotificationType.values,
            forwarding_last_update_at=None,
        )
        load_notification_entities(owner)
        load_notification_objects(owner)

        # when
        owner.send_new_notifications()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_forwarding_sync_fresh)
        notifications_processed = {
            int(args[1]["embeds"][0].footer.text[-10:])
            for args in mock_send_message.call_args_list
        }
        notifications_expected = set(
            owner.notification_set.filter(
                notif_type__in=NotificationType.values
            ).values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notifications_processed, notifications_expected)

    def test_should_only_send_selected_notification_types(self, mock_send_message):
        # given
        mock_send_message.return_value = 1
        selected_notif_types = [
            NotificationType.ORBITAL_ATTACKED,
            NotificationType.STRUCTURE_DESTROYED,
        ]
        webhook = WebhookFactory(notification_types=selected_notif_types)
        self.owner.webhooks.add(webhook)

        # when
        self.owner.send_new_notifications()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_forwarding_sync_fresh)
        notifications_processed = {
            int(args[1]["embeds"][0].footer.text[-10:])
            for args in mock_send_message.call_args_list
        }
        notifications_expected = set(
            Notification.objects.filter(
                notif_type__in=selected_notif_types
            ).values_list("notification_id", flat=True)
        )
        self.assertSetEqual(notifications_processed, notifications_expected)


class TestOwnerUpdateAssetEsi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.character = EveCharacterFactory()
        cls.user = UserMainDefaultOwnerFactory(main_character__character=cls.character)
        cls.owner = OwnerFactory(user=cls.user, assets_last_update_at=None)

    @pook.on
    def test_should_fetch_new_assets(self):
        # given
        structure = StructureFactory(owner=self.owner, quantum_core=False)
        service_module_type = CitadelServiceModuleTypeFactory()
        item_id = 1300000001001
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": item_id,
                    "location_flag": "ServiceSlot0",
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": service_module_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_assets_sync_fresh)
        obj = structure.items.get(pk=item_id)
        self.assertEqual(obj.eve_type_id, service_module_type.id)
        self.assertEqual(obj.location_flag, "ServiceSlot0")
        self.assertEqual(obj.quantity, 1)
        self.assertTrue(obj.is_singleton)

    @pook.on
    def test_should_updating_existing_assets(self):
        # given

        structure = StructureFactory(owner=self.owner, quantum_core=False)
        item = StructureItemFactory(structure=structure, is_singleton=False, quantity=3)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": item.is_singleton,
                    "item_id": item.id,
                    "location_flag": item.location_flag,
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 5,
                    "type_id": item.eve_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_assets_sync_fresh)
        item.refresh_from_db()
        self.assertEqual(item.quantity, 5)

    @pook.on
    def test_should_update_quantum_core(self):
        # given

        structure = StructureFactory(owner=self.owner, quantum_core=False)
        quantum_core_type = QuantumCoreTypeFactory()
        service_module_type = CitadelServiceModuleTypeFactory()
        item_1_id = 1300000001001
        item_2_id = 1300000002001
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": False,
                    "item_id": item_1_id,
                    "location_flag": "QuantumCoreRoom",
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": quantum_core_type.id,
                },
                {
                    "is_singleton": True,
                    "item_id": item_2_id,
                    "location_flag": "ServiceSlot0",
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": service_module_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_assets_sync_fresh)
        obj = structure.items.get(pk=item_1_id)
        self.assertEqual(obj.eve_type_id, quantum_core_type.id)
        self.assertEqual(
            obj.location_flag, StructureItem.LocationFlag.QUANTUM_CORE_ROOM
        )
        self.assertEqual(obj.quantity, 1)
        self.assertFalse(obj.is_singleton)

        obj = structure.items.get(pk=item_2_id)
        self.assertEqual(obj.eve_type_id, service_module_type.id)
        self.assertEqual(obj.location_flag, "ServiceSlot0")
        self.assertEqual(obj.quantity, 1)
        self.assertTrue(obj.is_singleton)

        structure.refresh_from_db()
        self.assertTrue(structure.has_fitting)
        self.assertTrue(structure.has_core)

    @patch(OWNERS_PATH + ".notify", spec=True)
    @pook.on
    def test_should_inform_user_about_successful_update(self, mock_notify):
        # given

        structure = StructureFactory(owner=self.owner, quantum_core=False)
        service_module_type = CitadelServiceModuleTypeFactory()
        item_2_id = 1300000002001
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": item_2_id,
                    "location_flag": "ServiceSlot0",
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": service_module_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi(user=self.user)

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_assets_sync_fresh)
        self.assertTrue(mock_notify.called)

    @pook.on
    def test_should_add_item_when_location_not_found(self):
        # given

        structure = StructureFactory(owner=self.owner, quantum_core=False)
        service_module_type = CitadelServiceModuleTypeFactory()
        item_id = 1300000001001
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": item_id,
                    "location_flag": "ServiceSlot0",
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": service_module_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=404,
            response_json={"error": "not found"},
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_assets_sync_fresh)
        obj = structure.items.get(pk=item_id)
        self.assertEqual(obj.eve_type_id, service_module_type.id)
        self.assertEqual(obj.location_flag, "ServiceSlot0")
        self.assertEqual(obj.quantity, 1)
        self.assertTrue(obj.is_singleton)

    @pook.on
    def test_should_raise_exception_when_esi_returns_http_error(self):
        # given

        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=500,
            response_json={"error": "some error"},
        )

        # when
        with self.assertRaises(HTTPServerError):
            self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertFalse(self.owner.is_assets_sync_fresh)

    @pook.on
    def test_should_remove_assets_that_no_longer_exist_for_existing_structure(self):
        # given

        structure = StructureFactory(owner=self.owner, quantum_core=False)
        service_module_type = CitadelServiceModuleTypeFactory()
        added_item_id = 1300000002001
        StructureItemFactory(structure=structure)  # should be removed
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": added_item_id,
                    "location_flag": "ServiceSlot0",
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1,
                    "type_id": service_module_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi()

        # then
        got = set(structure.items.values_list("id", flat=True))
        self.assertSetEqual(got, {added_item_id})

    @pook.on
    def test_should_handle_asset_moved_to_another_structure(self):
        # given

        structure_1 = StructureFactory(owner=self.owner, quantum_core=False)
        structure_2 = StructureFactory(owner=self.owner, quantum_core=False)
        item = StructureItemFactory(structure=structure_2)
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": item.id,
                    "location_flag": item.location_flag,
                    "location_id": structure_1.id,
                    "location_type": "item",
                    "quantity": item.quantity,
                    "type_id": item.eve_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.assertSetEqual(queryset_pks(structure_2.items.all()), set())
        self.assertSetEqual(queryset_pks(structure_1.items.all()), {item.id})

    @pook.on
    def test_should_remove_outdated_jump_fuel_alerts(self):
        # given

        structure = StructureFactory(owner=self.owner)
        config = JumpFuelAlertConfigFactory(threshold=100)
        structure.jump_fuel_alerts.create(structure=structure, config=config)

        fuel_type = LiquidOzoneTypeFactory()
        item_id = 1000000000004
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": False,
                    "item_id": item_id,
                    "location_flag": StructureItem.LocationFlag.STRUCTURE_FUEL,
                    "location_id": structure.id,
                    "location_type": "item",
                    "quantity": 1000,
                    "type_id": fuel_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[],
        )

        # when
        self.owner.update_asset_esi()
        # then
        self.assertEqual(structure.jump_fuel_alerts.count(), 0)

    # TODO: Add tests for error cases

    @pook.on
    def test_should_update_starbase_items(self):
        # given

        starbase = StarbaseFactory(owner=self.owner)  # position needs to match assets
        item_id = 1000000000004
        item_type = LiquidOzoneTypeFactory()
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": False,
                    "item_id": item_id,
                    "location_flag": StructureItem.LocationFlag.AUTOFIT,
                    "location_id": starbase.eve_solar_system.id,
                    "location_type": "solar_system",
                    "quantity": 3,
                    "type_id": item_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": item_id,
                    "position": {
                        "x": starbase.position_x,
                        "y": starbase.position_y,
                        "z": starbase.position_z,
                    },
                },
            ],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_assets_sync_fresh)
        self.assertSetEqual(queryset_pks(starbase.items.all()), {item_id})


@patch(OWNERS_PATH + ".STRUCTURES_FEATURE_SKYHOOKS", True)
@patch(OWNERS_PATH + ".EveSolarSystem.nearest_celestial")
class TestOwnerUpdateSkyhooks(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.character = EveCharacterFactory()
        cls.user = UserMainDefaultOwnerFactory(main_character__character=cls.character)
        cls.owner = OwnerFactory(user=cls.user, assets_last_update_at=None)
        cls.planet = EvePlanetFactory()

    @pook.on
    def test_should_create_new_skyhooks_from_scratch(self, mock_nearest_celestial):
        # given
        mock_nearest_celestial.return_value = NearestCelestial(
            eve_object=self.planet, distance=35_000_000, eve_type=self.planet.eve_type
        )
        skyhook_id = 1000000010001
        skyhook_type = SkyhookTypeFactory()
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": skyhook_id,
                    "location_flag": "AutoFit",
                    "location_id": self.planet.eve_solar_system.id,
                    "location_type": "solar_system",
                    "quantity": 1,
                    "type_id": skyhook_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": skyhook_id,
                    "position": PositionFactory(),
                },
            ],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertEqual(self.owner.structures.count(), 1)
        structure: Structure = self.owner.structures.get(pk=1000000010001)
        self.assertTrue(structure.is_skyhook)
        self.assertEqual(structure.eve_planet, self.planet)

    @pook.on
    def test_should_remove_obsolete_skyhooks(self, mock_nearest_celestial):
        # given
        mock_nearest_celestial.return_value = NearestCelestial(
            eve_object=self.planet, distance=35_000_000, eve_type=self.planet.eve_type
        )

        SkyhookFactory(owner=self.owner)  # should be removed
        skyhook = SkyhookFactory(owner=self.owner)  # should not be removed
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": skyhook.id,
                    "location_flag": StructureItem.LocationFlag.AUTOFIT,
                    "location_id": skyhook.eve_solar_system.id,
                    "location_type": "solar_system",
                    "quantity": 1,
                    "type_id": skyhook.eve_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": skyhook.id,
                    "position": PositionFactory(),
                },
            ],
        )

        # when
        self.owner.update_asset_esi()

        # then
        self.owner.refresh_from_db()
        self.assertEqual(queryset_pks(self.owner.structures.all()), {skyhook.id})

    @pook.on
    def test_should_handle_expected_errors_when_resolving_planet(
        self, mock_nearest_celestial
    ):
        # given
        mock_nearest_celestial.side_effect = ValueError

        skyhook_id = 1000000010001
        skyhook_type = SkyhookTypeFactory()
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": skyhook_id,
                    "location_flag": StructureItem.LocationFlag.AUTOFIT,
                    "location_id": self.planet.eve_solar_system.id,
                    "location_type": "solar_system",
                    "quantity": 1,
                    "type_id": skyhook_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": skyhook_id,
                    "position": PositionFactory(),
                },
            ],
        )
        # when
        self.owner.update_asset_esi()

        # then
        structure: Structure = self.owner.structures.first()
        self.assertEqual(structure.id, skyhook_id)
        self.assertTrue(structure.is_skyhook)
        self.assertIsNone(structure.eve_planet)

    @pook.on
    def test_should_ignore_no_reply_when_resolving_planet(self, mock_nearest_celestial):
        # given
        mock_nearest_celestial.return_value = None

        skyhook_id = 1000000010001
        skyhook_type = SkyhookTypeFactory()
        pook.get(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": skyhook_id,
                    "location_flag": StructureItem.LocationFlag.AUTOFIT,
                    "location_id": self.planet.eve_solar_system.id,
                    "location_type": "solar_system",
                    "quantity": 1,
                    "type_id": skyhook_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.character.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": skyhook_id,
                    "position": PositionFactory(),
                },
            ],
        )
        # when
        self.owner.update_asset_esi()

        # then
        structure: Structure = self.owner.structures.first()
        self.assertEqual(structure.id, skyhook_id)
        self.assertTrue(structure.is_skyhook)
        self.assertIsNone(structure.eve_planet)


class TestOwnerToken(TestCase):
    def test_should_return_valid_token(self):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        # when
        token = owner.characters.first().valid_token()
        # then
        self.assertIsInstance(token, Token)
        self.assertEqual(token.user, user)
        self.assertEqual(token.character_id, character.character_id)

    def test_should_return_none_if_no_valid_token_found(self):
        # given
        character = EveCharacterFactory()
        user = UserMainDefaultOwnerFactory(main_character__character=character)
        owner = OwnerFactory(user=user, characters=[character])
        user.token_set.first().scopes.clear()
        # when
        token = owner.characters.first().valid_token()
        # then
        self.assertIsNone(token)


@patch(OWNERS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", True)
@patch(OWNERS_PATH + ".notify_admins")
class TestOwnerUpdateIsUp(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.corporation = EveCorporationInfoFactory()

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", True)
    def test_should_do_nothing_when_still_up(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=True, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertTrue(result)
        self.assertFalse(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertTrue(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", False)
    def test_should_report_when_down(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=True, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertFalse(result)
        self.assertTrue(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertFalse(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", False)
    def test_should_not_report_again_when_still_down(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=False, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertFalse(result)
        self.assertFalse(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertFalse(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", True)
    def test_should_report_when_up_again(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=False, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertTrue(owner.is_up)

    @patch(OWNERS_PATH + ".Owner.are_all_syncs_ok", True)
    def test_should_report_when_up_for_the_first_time(self, mock_notify_admins):
        # given
        owner = OwnerFactory(
            corporation=self.corporation, is_up=None, is_alliance_main=True
        )
        # when
        result = owner.update_is_up()
        # then
        self.assertTrue(result)
        self.assertTrue(mock_notify_admins.called)
        owner.refresh_from_db()
        self.assertTrue(owner.is_up)



from unittest.mock import Mock, patch

from app_utils.django import app_labels
from app_utils.testing import NoSocketsTestCase

from structures.core import notification_timers
from structures.core.notification_types import NotificationType
from structures.models import Notification
from structures.tests.testdata.factories import (
    EveAllianceInfoFactory,
    EveCorporationInfoFactory,
    GeneratedNotificationFactory,
    OwnerFactory,
)
from structures.tests.testdata.helpers import (
    load_notification_entities,
    load_notification_objects,
)

MODULE_PATH = "structures.core.notification_timers"

if "timerboard" in app_labels():
    from allianceauth.timerboard.models import Timer as AuthTimer

    @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", False)
    @patch("allianceauth.timerboard.models.Timer", spec=True)
    def test_moon_timers_disabled(self, mock_Timer):
        # given
        notif = Notification.objects.get(notification_id=1000000404)
        # when
        result = notification_timers.add_or_remove_timer(notif)
        # then
        self.assertFalse(result)
        self.assertFalse(mock_Timer.objects.create.called)
        notif.refresh_from_db()
        self.assertFalse(notif.add_or_remove_timer())
        self.assertFalse(mock_Timer.delete.called)


if "structuretimers" in app_labels():

    @patch(
        "structuretimers.models._task_calc_timer_distances_for_all_staging_systems",
        Mock(),
    )
    @patch("structuretimers.models.STRUCTURETIMERS_NOTIFICATIONS_ENABLED", False)
    class TestNotificationAddToTimerboard(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            alliance = EveAllianceInfoFactory(alliance_id=3001)
            corporation = EveCorporationInfoFactory(
                corporation_id=2001, alliance=alliance
            )
            cls.owner = OwnerFactory(corporation=corporation, is_alliance_main=True)

            load_notification_entities(cls.owner)
            load_notification_objects(cls.owner)


if "timerboard" in app_labels() or "structuretimers" in app_labels():

    @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True)
    def test_normal(self):
        notification_without_timer_query = Notification.objects.filter(
            notification_id__in=[
                1000000401,
                1000000403,
                1000000405,
                1000000502,
                1000000503,
                1000000506,
                1000000507,
                1000000508,
                1000000509,
                1000000510,
                1000000511,
                1000000512,
                1000000513,
                1000000601,
                1000010509,
                1000010601,
            ]
        )
        for obj in notification_without_timer_query:
            self.assertFalse(obj.add_or_remove_timer())

        self.assertEqual(AuthTimer.objects.count(), 0)

        obj = Notification.objects.get(notification_id=1000000501)
        self.assertFalse(obj.add_or_remove_timer())

        obj = Notification.objects.get(notification_id=1000000504)
        self.assertTrue(obj.add_or_remove_timer())

        obj = Notification.objects.get(notification_id=1000000505)
        self.assertTrue(obj.add_or_remove_timer())

        obj = Notification.objects.get(notification_id=1000000602)
        self.assertTrue(obj.add_or_remove_timer())

        ids_set_1 = {x.id for x in AuthTimer.objects.all()}
        obj = Notification.objects.get(notification_id=1000000404)
        self.assertTrue(obj.add_or_remove_timer())

        self.assertEqual(AuthTimer.objects.count(), 4)

        # this should remove the right timer only
        obj = Notification.objects.get(notification_id=1000000402)
        obj.add_or_remove_timer()
        self.assertEqual(AuthTimer.objects.count(), 3)
        ids_set_2 = {x.id for x in AuthTimer.objects.all()}
        self.assertSetEqual(ids_set_1, ids_set_2)

    @patch(MODULE_PATH + ".STRUCTURES_MOON_EXTRACTION_TIMERS_ENABLED", True)
    def test_run_all(self):
        for obj in Notification.objects.all():
            timer_types = NotificationType.relevant_for_timerboard()
            with self.subTest(notif_type=obj.notif_type):
                is_timer = obj.notif_type in timer_types
                is_added = obj.add_or_remove_timer()
                self.assertEqual(is_timer, is_added)

    @patch(MODULE_PATH + ".STRUCTURES_TIMERS_ARE_CORP_RESTRICTED", False)
    def test_corp_restriction_1(self):
        # given
        notif = Notification.objects.get(notification_id=1000000504)
        # when
        result = notif.add_or_remove_timer()
        # then
        self.assertTrue(result)
        timer = AuthTimer.objects.first()
        self.assertFalse(timer.corp_timer)

    @patch(MODULE_PATH + ".STRUCTURES_TIMERS_ARE_CORP_RESTRICTED", True)
    def test_corp_restriction_2(self):
        obj = Notification.objects.get(notification_id=1000000504)
        self.assertTrue(obj.add_or_remove_timer())
        timer = AuthTimer.objects.first()
        self.assertTrue(timer.corp_timer)

    def test_timer_starbase_reinforcement(self):
        # given
        notif = GeneratedNotificationFactory()
        structure = notif.structures.first()
        # when
        result = notif.add_or_remove_timer()
        # then
        self.assertTrue(result)
        obj = AuthTimer.objects.first()
        self.assertEqual(obj.system, structure.eve_solar_system.name)
        self.assertEqual(obj.planet_moon, structure.eve_moon.name)
        self.assertEqual(obj.eve_time, structure.state_timer_end)
        self.assertEqual(obj.system, structure.eve_solar_system.name)
        self.assertEqual(obj.planet_moon, structure.eve_moon.name)
        self.assertEqual(obj.eve_time, structure.state_timer_end)
        self.assertEqual(obj.system, structure.eve_solar_system.name)
        self.assertEqual(obj.planet_moon, structure.eve_moon.name)
        self.assertEqual(obj.eve_time, structure.state_timer_end)



from typing import NamedTuple
from unittest.mock import patch

import pook

from django.test import TestCase
from eveuniverse.tests.testdata.factories_2 import EvePlanetFactory

from structures.constants import EveCorporationId
from structures.models import PocoDetails, Structure, owners
from structures.tests.testdata.factories import (
    CustomsOfficeFactory,
    CustomsOfficeTypeFactory,
    EveEntityCorporationFactory,
    OwnerFactory,
    PositionFactory,
    UserMainDefaultOwnerFactory,
)

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
@patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
class TestUpdateCustomOfficesEsi(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = UserMainDefaultOwnerFactory()
        cls.owner = OwnerFactory(user=cls.user, structures_last_update_at=None)
        cls.corporation_id = cls.owner.corporation.corporation_id
        EveEntityCorporationFactory(
            id=EveCorporationId.DED, name="DED"
        )  # for notifications

    @pook.on
    def test_can_sync_pocos_and_remove_old(self):
        # given
        planet = EvePlanetFactory(
            name="Amamake V",
            eve_solar_system__enabled_sections=1,  # 1 = Planets
            eve_type__name="Planet (Barren)",
        )
        poco_type = CustomsOfficeTypeFactory()
        customs_office_id = 1200000000003
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": customs_office_id,
                    "position": PositionFactory(),
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=200,
            response_json=[
                {
                    "item_id": customs_office_id,
                    "name": f"Customs Office ({planet.name})",
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/customs_offices",
            reply=200,
            response_json=[
                {
                    "alliance_tax_rate": 0.02,
                    "allow_access_with_standings": True,
                    "allow_alliance_access": True,
                    "bad_standing_tax_rate": 0.3,
                    "corporation_tax_rate": 0.02,
                    "excellent_standing_tax_rate": 0.02,
                    "good_standing_tax_rate": 0.02,
                    "neutral_standing_tax_rate": 0.02,
                    "office_id": customs_office_id,
                    "reinforce_exit_end": 21,
                    "reinforce_exit_start": 19,
                    "standing_level": "terrible",
                    "system_id": planet.eve_solar_system.id,
                    "terrible_standing_tax_rate": 0.5,
                }
            ],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        CustomsOfficeFactory(owner=owner, id=1200000000010, name="delete-me")

        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        # must contain all expected structures
        self.assertSetEqual(owner.structures.ids(), {customs_office_id})
        self.assertSetEqual(
            set(PocoDetails.objects.values_list("structure_id", flat=True)),
            {customs_office_id},
        )

        # verify attributes for POCO
        structure = Structure.objects.get(id=customs_office_id)
        self.assertEqual(structure.name, "Planet (Barren)")
        self.assertEqual(structure.eve_solar_system_id, planet.eve_solar_system.id)
        self.assertEqual(
            int(structure.owner.corporation.corporation_id), self.corporation_id
        )
        self.assertEqual(structure.eve_type_id, poco_type.id)
        self.assertEqual(structure.reinforce_hour, 20)
        self.assertEqual(structure.state, Structure.State.UNKNOWN)
        self.assertEqual(structure.eve_planet_id, planet.id)

        # verify attributes for POCO details
        details: PocoDetails = structure.poco_details
        self.assertEqual(details.alliance_tax_rate, 0.02)
        self.assertTrue(details.allow_access_with_standings)
        self.assertTrue(details.allow_alliance_access)
        self.assertEqual(details.bad_standing_tax_rate, 0.3)
        self.assertEqual(details.corporation_tax_rate, 0.02)
        self.assertEqual(details.excellent_standing_tax_rate, 0.02)
        self.assertEqual(details.good_standing_tax_rate, 0.02)
        self.assertEqual(details.neutral_standing_tax_rate, 0.02)
        self.assertEqual(details.reinforce_exit_end, 21)
        self.assertEqual(details.reinforce_exit_start, 19)
        self.assertEqual(details.standing_level, PocoDetails.StandingLevel.TERRIBLE)
        self.assertEqual(details.terrible_standing_tax_rate, 0.5)

    @pook.on
    def test_poco_has_no_name_when_no_asset_data(self):
        # given
        planet = EvePlanetFactory(
            name="Amamake V",
            eve_solar_system__name="Amamake",
            eve_solar_system__enabled_sections=1,  # 1 = Planets
            eve_type__name="Planet (Barren)",
        )
        CustomsOfficeTypeFactory()
        customs_office_id = 1200000000003
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": customs_office_id,
                    "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/customs_offices",
            reply=200,
            response_json=[
                {
                    "alliance_tax_rate": 0.02,
                    "allow_access_with_standings": True,
                    "allow_alliance_access": True,
                    "bad_standing_tax_rate": 0.3,
                    "corporation_tax_rate": 0.02,
                    "excellent_standing_tax_rate": 0.02,
                    "good_standing_tax_rate": 0.02,
                    "neutral_standing_tax_rate": 0.02,
                    "office_id": customs_office_id,
                    "reinforce_exit_end": 21,
                    "reinforce_exit_start": 19,
                    "standing_level": "terrible",
                    "system_id": planet.eve_solar_system.id,
                    "terrible_standing_tax_rate": 0.5,
                }
            ],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        # when
        owner.update_structures_esi()

        # then
        self.assertSetEqual(owner.structures.ids(), {customs_office_id})
        structure = Structure.objects.get(id=customs_office_id)
        self.assertEqual(structure.name, "")
        self.assertEqual(structure.eve_solar_system_id, planet.eve_solar_system.id)

    @pook.on
    def test_should_not_break_or_delete_on_http_error_when_fetching_custom_offices(
        self,
    ):
        # given
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/customs_offices",
            reply=500,
            response_json={"error": "internal error"},
        )
        customs_office_id = 1200000000003
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        CustomsOfficeFactory(owner=owner, id=customs_office_id, name="keep-me")
        # when
        owner.update_structures_esi()
        # then
        self.assertSetEqual(owner.structures.ids(), {customs_office_id})

    @pook.on
    def test_should_not_break_when_name_not_found(self):
        # given
        planet = EvePlanetFactory(
            name="Amamake V",
            eve_solar_system__name="Amamake",
            eve_solar_system__enabled_sections=1,  # 1 = Planets
            eve_type__name="Planet (Barren)",
        )
        CustomsOfficeTypeFactory()
        customs_office_id = 1200000000003
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": customs_office_id,
                    "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=404,
            response_json={"error": "not found"},
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/customs_offices",
            reply=200,
            response_json=[
                {
                    "alliance_tax_rate": 0.02,
                    "allow_access_with_standings": True,
                    "allow_alliance_access": True,
                    "bad_standing_tax_rate": 0.3,
                    "corporation_tax_rate": 0.02,
                    "excellent_standing_tax_rate": 0.02,
                    "good_standing_tax_rate": 0.02,
                    "neutral_standing_tax_rate": 0.02,
                    "office_id": customs_office_id,
                    "reinforce_exit_end": 21,
                    "reinforce_exit_start": 19,
                    "standing_level": "terrible",
                    "system_id": planet.eve_solar_system.id,
                    "terrible_standing_tax_rate": 0.5,
                }
            ],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)

        # when
        owner.update_structures_esi()

        # then
        self.assertSetEqual(owner.structures.ids(), {customs_office_id})
        structure = Structure.objects.get(id=customs_office_id)
        self.assertEqual(structure.name, "")
        self.assertEqual(structure.eve_solar_system_id, planet.eve_solar_system.id)

    @pook.on
    def test_should_have_empty_name_when_no_match_with_planet(self):
        # given
        planet = EvePlanetFactory(
            name="Amamake V",
            eve_solar_system__name="Amamake",
            eve_solar_system__enabled_sections=1,  # 1 = Planets
            eve_type__name="Planet (Barren)",
        )
        CustomsOfficeTypeFactory()
        customs_office_id = 1200000000003
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/locations",
            reply=200,
            response_json=[
                {
                    "item_id": customs_office_id,
                    "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/assets/names",
            reply=200,
            response_json=[
                {
                    "item_id": customs_office_id,
                    "name": "invalid name",
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{self.corporation_id}/customs_offices",
            reply=200,
            response_json=[
                {
                    "alliance_tax_rate": 0.02,
                    "allow_access_with_standings": True,
                    "allow_alliance_access": True,
                    "bad_standing_tax_rate": 0.3,
                    "corporation_tax_rate": 0.02,
                    "excellent_standing_tax_rate": 0.02,
                    "good_standing_tax_rate": 0.02,
                    "neutral_standing_tax_rate": 0.02,
                    "office_id": customs_office_id,
                    "reinforce_exit_end": 21,
                    "reinforce_exit_start": 19,
                    "standing_level": "terrible",
                    "system_id": planet.eve_solar_system.id,
                    "terrible_standing_tax_rate": 0.5,
                }
            ],
        )
        owner = OwnerFactory(user=self.user, structures_last_update_at=None)
        CustomsOfficeFactory(owner=owner, id=1200000000010, name="delete-me")

        # when
        owner.update_structures_esi()

        # then
        self.assertSetEqual(owner.structures.ids(), {customs_office_id})
        structure = Structure.objects.get(id=customs_office_id)
        self.assertEqual(structure.name, "")


class TestExtractPlanetNameFromAsset(TestCase):
    def test_ok(self):
        class Case(NamedTuple):
            input: str
            want: str

        cases = [
            Case("Customs Office (Amamake V)", "Amamake V"),
            Case("Customs Office (1-PGSG VI)", "1-PGSG VI"),
            Case("Customs Office (1-PGSG VII)", "1-PGSG VII"),
            Case("invalid name", ""),
        ]
        for tc in cases:
            with self.subTest(input=tc.input):
                got = owners._extract_planet_name_from_asset(tc.input)
                self.assertEqual(got, tc.want)



import datetime as dt

import pook

from django.test import TestCase, override_settings
from django.utils.timezone import now
from eveuniverse.models import EvePlanet, EveSolarSystem
from eveuniverse.tests.testdata.factories_2 import EveMoonFactory, EveTypeFactory

from app_utils.django import app_labels
from app_utils.testing import queryset_pks, reset_celery_once_locks

from structures import tasks
from structures.core.notification_types import NotificationType
from structures.tests.testdata.factories import (
    CitadelTypeFactory,
    CustomsOfficeTypeFactory,
    EveEntityAllianceFactory,
    EveEntityCorporationFactory,
    FuelBlockTypeFactory,
    NotificationFactory,
    OwnerFactory,
    PositionFactory,
    RawNotificationFactory,
    SkyhookTypeFactory,
    StarbaseFactory,
    StarbaseTypeFactory,
    StructureFactory,
    WebhookFactory,
)

if "structuretimers" in app_labels():
    from structuretimers.models import Timer as StructureTimer
else:
    StructureTimer = None

if "timerboard" in app_labels():
    from allianceauth.timerboard.models import Timer as AuthTimer
else:
    AuthTimer = None

MANAGERS_PATH = "structures.managers"
OWNERS_PATH = "structures.models.owners"
NOTIFICATIONS_PATH = "structures.models.notifications"
TASKS_PATH = "structures.tasks"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestTasks(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        reset_celery_once_locks("structures")

    @pook.on
    def test_should_fetch_new_structures_from_esi(self):
        # given
        owner = OwnerFactory()
        corporation_id = owner.corporation.corporation_id
        moon = EveMoonFactory(eve_planet__eve_solar_system__enabled_sections=1)
        planet: EvePlanet = moon.eve_planet
        solar_system: EveSolarSystem = moon.eve_planet.eve_solar_system

        structure_id = 1000000000001
        structure_type = CitadelTypeFactory()

        customs_office_id = 1200000000003
        CustomsOfficeTypeFactory()

        starbase_id = 1300000000001
        fuel_1_type = EveTypeFactory()
        fuel_2_type = EveTypeFactory()
        moon = EveMoonFactory()
        starbase_type = StarbaseTypeFactory()

        skyhook_id = 1000000010001
        skyhook_type = SkyhookTypeFactory()

        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets",
            reply=200,
            response_json=[
                {
                    "is_singleton": True,
                    "item_id": skyhook_id,
                    "location_flag": "AutoFit",
                    "location_id": planet.eve_solar_system.id,
                    "location_type": "solar_system",
                    "quantity": 1,
                    "type_id": skyhook_type.id,
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/locations",
            reply=200,
            json=[customs_office_id],
            response_json=[
                {
                    "item_id": customs_office_id,
                    "position": PositionFactory(),
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/locations",
            reply=200,
            json=[starbase_id],
            response_json=[
                {
                    "item_id": starbase_id,
                    "position": PositionFactory(),
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/locations",
            reply=200,
            json=[skyhook_id],
            response_json=[
                {
                    "item_id": skyhook_id,
                    "position": PositionFactory(),
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/names",
            reply=200,
            json=[starbase_id],
            response_json=[
                {"item_id": starbase_id, "name": "Home Sweat Home"},
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/names",
            reply=200,
            json=[customs_office_id],
            response_json=[
                {
                    "item_id": customs_office_id,
                    "name": f"Customs Office ({planet.name})",
                },
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/customs_offices",
            reply=200,
            response_json=[
                {
                    "alliance_tax_rate": 0.02,
                    "allow_access_with_standings": True,
                    "allow_alliance_access": True,
                    "bad_standing_tax_rate": 0.3,
                    "corporation_tax_rate": 0.02,
                    "excellent_standing_tax_rate": 0.02,
                    "good_standing_tax_rate": 0.02,
                    "neutral_standing_tax_rate": 0.02,
                    "office_id": customs_office_id,
                    "reinforce_exit_end": 21,
                    "reinforce_exit_start": 19,
                    "standing_level": "terrible",
                    "system_id": planet.eve_solar_system.id,
                    "terrible_standing_tax_rate": 0.5,
                }
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/starbases",
            reply=200,
            response_json=[
                {
                    "moon_id": moon.id,
                    "starbase_id": starbase_id,
                    "state": "online",
                    "system_id": moon.eve_planet.eve_solar_system.id,
                    "type_id": starbase_type.id,
                }
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/starbases/{starbase_id}",
            reply=200,
            response_json={
                "allow_alliance_members": True,
                "allow_corporation_members": True,
                "anchor": "config_starbase_equipment_role",
                "attack_if_at_war": False,
                "attack_if_other_security_status_dropping": False,
                "fuel_bay_take": "config_starbase_equipment_role",
                "fuel_bay_view": "starbase_fuel_technician_role",
                "fuels": [
                    {
                        "quantity": 960,
                        "type_id": fuel_1_type.id,
                    },
                    {
                        "quantity": 11678,
                        "type_id": fuel_2_type.id,
                    },
                ],
                "offline": "config_starbase_equipment_role",
                "online": "config_starbase_equipment_role",
                "unanchor": "config_starbase_equipment_role",
                "use_alliance_standings": True,
            },
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/structures",
            reply=200,
            response_json=[
                {
                    "corporation_id": corporation_id,
                    "profile_id": 101853,
                    "reinforce_hour": 18,
                    "state": "shield_vulnerable",
                    "structure_id": structure_id,
                    "system_id": solar_system.id,
                    "type_id": structure_type.id,
                },
            ],
        )
        pook.get(
            "https://esi.evetech.net/sovereignty/map",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/universe/structures/{structure_id}",
            reply=200,
            response_json={
                "owner_id": corporation_id,
                "name": f"{solar_system.name} - Test Structure Alpha",
                "position": {
                    "x": 55028384780.0,
                    "y": 7310316270.0,
                    "z": -163686684205.0,
                },
                "solar_system_id": solar_system.id,
                "type_id": structure_type.id,
            },
        )
        pook.get(
            f"https://evesdeapi.kalkoken.net/latest/universe/systems/{solar_system.id}/nearest_celestials",
            reply=200,
            response_json=[
                {
                    "distance": 10,
                    "group_id": 7,
                    "group_name": "Planet",
                    "item_id": planet.id,
                    "name": planet.name,
                    "position": PositionFactory(),
                    "type_id": planet.eve_type.id,
                    "type_name": planet.eve_type.name,
                },
            ],
        )

        # when
        tasks.update_all_structures.delay()

        # then
        got = queryset_pks(owner.structures.all())
        want = {
            customs_office_id,
            skyhook_id,
            starbase_id,
            structure_id,
        }
        self.assertSetEqual(got, want)
        self.assertTrue(pook.isdone(), msg=pook.pending_mocks())

    @pook.on
    def test_should_send_notification_and_create_timers_for_reinforced_starbase(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.TOWER_REINFORCED_EXTRA]
        )
        owner = OwnerFactory(webhooks=[webhook])
        corporation_id = owner.corporation.corporation_id
        eve_character = owner.characters.first().character_ownership.character

        starbase = StarbaseFactory(owner=owner)
        reinforced_until = now() + dt.timedelta(hours=12)
        fuel_type = FuelBlockTypeFactory()

        discord_mock = pook.post(webhook.url, reply=204)
        pook.get(
            f"https://esi.evetech.net/characters/{eve_character.character_id}/notifications",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets",
            reply=200,
            response_json=[],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/locations",
            reply=200,
            json=[starbase.id],
            response_json=[
                {
                    "item_id": starbase.id,
                    "position": PositionFactory(),
                },
            ],
        )
        pook.post(
            f"https://esi.evetech.net/corporations/{corporation_id}/assets/names",
            reply=200,
            json=[starbase.id],
            response_json=[
                {"item_id": starbase.id, "name": starbase.name},
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/customs_offices",
            reply=200,
            response_json=[],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/starbases",
            reply=200,
            response_json=[
                {
                    "moon_id": starbase.eve_moon.id,
                    "starbase_id": starbase.id,
                    "state": "reinforced",
                    "system_id": starbase.eve_moon.eve_planet.eve_solar_system.id,
                    "type_id": starbase.eve_type.id,
                    "reinforced_until": reinforced_until.isoformat(),
                }
            ],
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/starbases/{starbase.id}",
            reply=200,
            response_json={
                "allow_alliance_members": True,
                "allow_corporation_members": True,
                "anchor": "config_starbase_equipment_role",
                "attack_if_at_war": False,
                "attack_if_other_security_status_dropping": False,
                "fuel_bay_take": "config_starbase_equipment_role",
                "fuel_bay_view": "starbase_fuel_technician_role",
                "fuels": [
                    {
                        "quantity": 960,
                        "type_id": fuel_type.id,
                    },
                ],
                "offline": "config_starbase_equipment_role",
                "online": "config_starbase_equipment_role",
                "unanchor": "config_starbase_equipment_role",
                "use_alliance_standings": True,
            },
        )
        pook.get(
            f"https://esi.evetech.net/corporations/{corporation_id}/structures",
            reply=200,
            response_json=[],
        )
        pook.get(
            "https://esi.evetech.net/sovereignty/map",
            reply=200,
            response_json=[],
        )

        # when
        tasks.update_all_structures.delay()
        tasks.fetch_all_notifications.delay()

        # then
        self.assertEqual(discord_mock.calls, 1)
        self.assertIn(
            starbase.name, discord_mock.matches[0].json["embeds"][0]["description"]
        )

        if StructureTimer:
            self.assertTrue(StructureTimer.objects.exists())

        if AuthTimer:
            self.assertTrue(AuthTimer.objects.exists())

        self.assertTrue(pook.isdone(), msg=pook.pending_mocks())

    @pook.on
    def test_should_fetch_and_send_notification_when_enabled_for_webhook(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE]
        )
        owner = OwnerFactory(webhooks=[webhook], is_alliance_main=True)
        eve_character = owner.characters.first().character_ownership.character
        notif = RawNotificationFactory()

        discord_mock = pook.post(webhook.url, reply=204)
        pook.get(
            f"https://esi.evetech.net/characters/{eve_character.character_id}/notifications",
            reply=200,
            response_json=[notif],
        )

        # when
        tasks.fetch_all_notifications.delay()

        # then
        self.assertEqual(discord_mock.calls, 1)
        self.assertIn(
            "now eligible", discord_mock.matches[0].json["embeds"][0]["description"]
        )
        self.assertTrue(pook.isdone(), msg=pook.pending_mocks())

    @pook.on
    def test_should_fetch_and_send_notification_when_enabled_for_webhook_all_anchoring(
        self,
    ):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.SOV_ALL_ANCHORING_MSG]
        )
        owner = OwnerFactory(webhooks=[webhook], is_alliance_main=False)
        eve_character = owner.characters.first().character_ownership.character
        alliance = EveEntityAllianceFactory(
            id=owner.corporation.alliance_id,
            name=owner.corporation.alliance.alliance_name,
        )
        corporation = EveEntityCorporationFactory(
            id=owner.corporation.corporation_id, name=owner.corporation.corporation_name
        )
        starbase = StarbaseFactory(owner=owner)
        notif = RawNotificationFactory(
            type="AllAnchoringMsg",
            sender=corporation,
            data={
                "allianceID": alliance.id,
                "corpID": corporation.id,
                "corpsPresent": [{"allianceID": alliance.id, "corpID": corporation.id}],
                "moonID": starbase.eve_moon.id,
                "solarSystemID": starbase.eve_solar_system.id,
                "towers": [
                    {"moonID": starbase.eve_moon.id, "typeID": starbase.eve_type.id}
                ],
                "typeID": starbase.eve_type.id,
            },
        )

        discord_mock = pook.post(webhook.url, reply=204)
        pook.get(
            f"https://esi.evetech.net/characters/{eve_character.character_id}/notifications",
            reply=200,
            response_json=[notif],
        )

        # when
        tasks.fetch_all_notifications.delay()

        # then
        self.assertEqual(discord_mock.calls, 1)
        self.assertIn(
            "has anchored in", discord_mock.matches[0].json["embeds"][0]["description"]
        )
        self.assertTrue(pook.isdone(), msg=pook.pending_mocks())

    @pook.on
    def test_should_fetch_new_notification_and_send_to_webhook_and_create_timers(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.STRUCTURE_LOST_SHIELD]
        )
        owner = OwnerFactory(webhooks=[webhook])
        eve_character = owner.characters.first().character_ownership.character
        structure = StructureFactory(owner=owner)

        discord_mock = pook.post(webhook.url, reply=204)
        pook.get(
            f"https://esi.evetech.net/characters/{eve_character.character_id}/notifications",
            reply=200,
            response_json=[
                RawNotificationFactory(
                    type="StructureLostShields",
                    data={
                        "solarsystemID": structure.eve_solar_system.id,
                        "structureID": structure.id,
                        "structureShowInfoData": [
                            "showinfo",
                            structure.eve_type.id,
                            structure.id,
                        ],
                        "structureTypeID": structure.eve_type.id,
                        "timeLeft": 3432362784823,
                        "timestamp": 132977978640000000,
                        "vulnerableTime": 9000000000,
                    },
                ),
            ],
        )

        # when
        tasks.fetch_all_notifications.delay()

        # then
        self.assertEqual(discord_mock.calls, 1)
        self.assertIn(
            structure.name, discord_mock.matches[0].json["embeds"][0]["description"]
        )

        if StructureTimer:
            obj = StructureTimer.objects.first()
            self.assertEqual(obj.eve_solar_system.id, structure.eve_solar_system.id)

        if AuthTimer:
            obj = AuthTimer.objects.first()
            self.assertEqual(obj.system, structure.eve_solar_system.name)

        self.assertTrue(pook.isdone(), msg=pook.pending_mocks())

    @pook.on
    def test_should_send_selected_notif_types_only(self):
        # given
        webhook = WebhookFactory(
            notification_types=[NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE]
        )
        owner = OwnerFactory(webhooks=[webhook], is_alliance_main=True)
        eve_character = owner.characters.first().character_ownership.character
        NotificationFactory(
            owner=owner, notif_type=NotificationType.WAR_CORPORATION_NO_LONGER_ELIGIBLE
        )
        NotificationFactory(
            owner=owner, notif_type=NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE
        )

        discord_mock = pook.post(webhook.url, reply=204)
        pook.get(
            f"https://esi.evetech.net/characters/{eve_character.character_id}/notifications",
            reply=200,
            response_json=[],
        )

        # when
        tasks.fetch_all_notifications.delay()

        # then
        self.assertEqual(discord_mock.calls, 1)
        self.assertIn(
            "war declarations", discord_mock.matches[0].json["embeds"][0]["title"]
        )
        self.assertTrue(pook.isdone(), msg=pook.pending_mocks())
